#ifndef lint
static char rcsid[] = "$Header: /net/hopkins/var/home/tristan/mpeg/dec/RCS/mtv.c,v 1.182 1997/08/29 04:11:47 tristan Exp $";
static char copyright[] = "Copyright (c) 1997, MpegTV LLC";
#endif


/*
 * Disclaimer of Warranty
 *
 * These software programs are available to the user without any license fee or
 * royalty on an "as is" basis.  Copyright-holder disclaims
 * any and all warranties, whether express, implied, or statuary, including any
 * implied warranties or merchantability or of fitness for a particular
 * purpose.  In no event shall the copyright-holder be liable for any
 * incidental, punitive, or consequential damages of any kind whatsoever
 * arising from the use of these programs.
 *
 * This disclaimer of warranty extends to the user of these programs and user's
 * customers, employees, agents, transferees, successors, and assigns.
 *
 */

#include "misc.h"
#include "pmp.h"
#include "msg.h"
#include "constrained.h"
#include "dither_api.h"
#include "state_api.h"
#include "audio_api.h"
#include "sync_api.h"
#include "mode_api.h"
#include "forms.h"
#include "controls.h"
#include "exit.h"
#include "version.h"
#include "g2a.h"
#include "get_time_now.h"

#ifndef DEBUG
#	define GET_TIME_NOW_INLINE	/* use in-line version if possible - not implemented in public version of get_time_now.h */
#endif
#include "get_time_now.h"

#include "mtv.h"

#if (FL_VERSION != 0) || ((FL_REVISION != 86) && (FL_REVISION != 81))
#	error unsupported XForms version - must be 0.81 or 0.86
#endif

/* gui_challenge.h not available in public distribution */
#ifdef CHALLENGE
#	include "gui_challenge.h"
#endif

/*--------------------------------------------------------------------------+
| Configuration																|
+--------------------------------------------------------------------------*/

/*
 * Initial PCM volume setting at startup.
 * PCM volume is set to only 75% because settings to 100% would cause
 * audio to saturate on some systems.
 */
#define INITIAL_PCM_VOLUME	(-1)
#define INITIAL_PCM_BALANCE	(-1)
#define INITIAL_MASTER_VOLUME	1.0
#define INITIAL_MASTER_BALANCE	0.5

/*--------------------------------------------------------------------------+
| Macros																	|
+--------------------------------------------------------------------------*/

/* fast forward plays FAST_FORWARD_COEF times faster */
#define FAST_FORWARD_COEF	8

/* frequence of slider position update when playing (in sec) */
#define SLIDER_UPDATE_FREQUENCY	1

#define G2A_SEND_VOID(com, flags)		g2a_send_void(&Player, com, flags)
#define G2A_SEND_INT(com, flags, n)		g2a_send_int(&Player, com, flags, n)
#define G2A_SEND_DOUBLE(com, flags, n)	g2a_send_double(&Player, com, flags, n)
#define G2A_SEND_STRING(com, flags, n)	g2a_send_string(&Player, com, flags, n)

/* values for slider controls */
#define GAMMA_INVALID	(-1)
#ifdef SGI
#define GAMMA_DEFAULT	0.8
#elif defined(SOLARIS)
#define GAMMA_DEFAULT	1.4
#else
#define GAMMA_DEFAULT	1.0
#endif

#define GAMMA_MIN	0.1
#define GAMMA_MAX	2.0

#define CONTRAST_INVALID	(-1)
#define CONTRAST_DEFAULT	/*(36.0/219.0)*/ 0.0
#define CONTRAST_MIN	(-0.50)
#define CONTRAST_MAX	0.50

#define BRIGHTNESS_INVALID	(-1)
#define BRIGHTNESS_DEFAULT	/*(20.0/219.0)*/ 0.0
#define BRIGHTNESS_MIN	(-0.50)
#define BRIGHTNESS_MAX	0.50

#define FRAMERATE_MIN	1.0
#define FRAMERATE_MAX	60.0

#define SET_ONELINER(ob, str) STMT( \
		(ob)->u_vdata = (str); \
		fl_set_object_posthandler(ob, post_handler); \
	)

/*--------------------------------------------------------------------------+
| Data types																|
+--------------------------------------------------------------------------*/
/* data type for user options */
typedef struct {
	Bool	loop;
	ZOOM_TYPE	zoom;
	Bool	iconify;
	Bool	auto_exit;
	Bool	auto_play;
	DITHER_TYPE	dither_type;
	SYNC_TYPE	sync_type;
	double	framerate;
	const char *filename;
	Bool	stdin_mode;
	int		port;
	int		port_udp;
	u_int		msg_select;
	double	gamma;
	double	contrast;
	double	brightness;
	u_int		audio_msg_select;
	AUDIO_MODE	audio_mode;
	AUDIO_QUALITY	audio_quality;
	AUDIO_CHANNEL	audio_channel;
	AUDIO_CONTROLS	audio_controls;
	double		audio_pcm_volume;
	double		audio_pcm_balance;
	double		audio_master_volume;
	double		audio_master_balance;
	int			audio_ports;
	const char **xforms_argv;
	int		pmp_base_fd;
	int		existing_window;	/* -1 means not set by option -W<n> */
	off_t	loop_start_pos;
	off_t	loop_end_pos;
	int	mtvp_geometry_x;
	int	mtvp_geometry_y;
	int	geometry_x;
	int	geometry_y;
	u_int	geometry_w;
	u_int	geometry_h;
} USER_OPTIONS;

/* for the file selector state */
typedef struct {
	char directory[PATH_MAX];
	char pattern[PATH_MAX];
	char filename[PATH_MAX];
} SEL_STATE;

/* File menu items */
typedef enum {
	FILE_MENU_NONE = -1,
	FILE_MENU_HANGING = 0,	/* undocumented feature of Xforms menus */
	FILE_MENU_OPEN = 1,
	FILE_MENU_STDIN,
	FILE_MENU_CLOSE,
	FILE_MENU_SAVE_AS,
	FILE_MENU_INFOS,
	FILE_MENU_QUIT
} FILE_MENU;

/* Options menu items */
typedef enum {
	OPTIONS_MENU_NONE = -1,
	OPTIONS_MENU_HANGING = 0,	/* undocumented feature of Xforms menus */
	OPTIONS_MENU_AUDIO = 1,
	OPTIONS_MENU_VIDEO,
	OPTIONS_MENU_SYNC
} OPTIONS_MENU;

/* Window menu items */
typedef enum {
	WINDOW_MENU_NONE = -1,
	WINDOW_MENU_HANGING = 0,	/* undocumented feature of Xforms menus */
	WINDOW_MENU_RAISE = 1,
	WINDOW_MENU_FRAME_INFOS,
	WINDOW_MENU_SAVE_PPM,
	WINDOW_MENU_SAVE_TGA
} WINDOW_MENU;

/* Help menu items */
typedef enum {
	HELP_MENU_NONE = -1,
	HELP_MENU_HANGING = 0,	/* undocumented feature of Xforms menus */
	HELP_MENU_ABOUT = 1,
	HELP_MENU_LICENSE = 2
} HELP_MENU;

/* Slider control */
typedef struct {
	Bool	not_ready;
	double	pending;
} SLIDER_CONTROL;

/*--------------------------------------------------------------------------+
| Globals																	|
+--------------------------------------------------------------------------*/

/* The form pointers */
FD_controls *Controls;
FD_about *About;
FD_license *License;
FD_audio *Audio;
FD_sync *Sync;
FD_infos *Infos;
FD_frame_infos *Frame_infos;
FD_video *Video;

/* to communicate with player */
PMP_STRUCT	Player;

/*--------------------------------------------------------------------------+
| Statics																	|
+--------------------------------------------------------------------------*/
/* name of run-time Player, if not "mtvp" */
static const char *Player_name;

/* default for the file selectors */
static SEL_STATE Open_file = {"", "*.mpg", ""};
static SEL_STATE Save_file = {"", "*.mpg", ""};
static SEL_STATE Save_image_ppm = {"", "*.ppm", ""};
static SEL_STATE Save_image_tga = {"", "*.tga", ""};

/* for the play/pause button */
static STATE Gui_play_pause_set_state = STATE_PLAYING;

/* player's state, used by the my_idle() routine */
static STATE Gui_state = STATE_STOPPED;

/* user options */
static USER_OPTIONS User_options;


/* The pid of the control process */
static pid_t Pid;

/*
 * set Not_ready when sending a seek command. Reset it when a A2G_COM_DISPLAY_READY
 * message id received.
 * Not_ready is used to avoid piling up seek commands faster than the
 * player can process them.
 */
static Bool Not_ready;
static int Pending_seek = -1;	/* pending seek or -1 */

static double Nominal_framerate;		/* nominal framerate */
static Bool Framerate_is_nominal;	/* True means mtvp is playing in sync at Nominal_framerate */
static Bool Specified_framerate_is_nominal;
static int Bits_filesize = -1;	/* stream size (in bytes) or -1 if n/a */
static u_int Mux_bitrate;			/* mux bitrate or 0 if n/a */
static u_int Video_bitrate;		/* video bitrate or 0 if n/a */

static SLIDER_CONTROL Gamma;		/* for gamma slider */
static SLIDER_CONTROL Contrast;	/* for contrast slider */
static SLIDER_CONTROL Brightness;	/* for brightness slider */

static Pixmap Icon_pixmap;			/* MpegTV Icon for closed windows */

static FL_OBJECT *Oneliner_object;	/* NULL or ptr to current object for oneliner */

static off_t Stream_pos;		/* last position received from player */

/* Fast_forward is necessary to work around a bug of the IN-OUT buttons in Xforms! */
static Bool Fast_forward;		/* if True we are playing in FFW mode */
static STATE State_before_fast_forward;/* restore this state after FFW */
static Bool Seeking;			/* if True we are playing in seeking mode */
static STATE State_before_seeking;		/* restore this state after seeking */

static int Audio_ports;			/* Audio ports available */

static const char *License_lines[] = {
#include "player-license.h"
};

/*--------------------------------------------------------------------------+
| get_elapsed_seconds														|
+--------------------------------------------------------------------------*/
static INLINE int
get_elapsed_seconds(pos)
	off_t		pos;
{
#if 0 /* Tristan Savatier - Sun Aug 3 02:45:29 PDT 1997 */
//	if (Bits_filesize == -1) {
//		return -1;
//	}
#endif /* 0 - Tristan Savatier - Sun Aug 3 02:45:29 PDT 1997 */
	if (Mux_bitrate / BITS(char) != 0) {
		return (int)((u_int)pos / (Mux_bitrate / BITS(char)));
	} else if (Video_bitrate / BITS(char) != 0) {
		return (int)((u_int)pos / (Video_bitrate / BITS(char)));
	} else {
		return -1;
	}
}

/*--------------------------------------------------------------------------+
| get_time_string															|
+--------------------------------------------------------------------------*/
static const char *
get_time_string(sec)
	int		sec;
{
	static char buffer[9];

#if 0 /* Tristan Savatier - Sat Aug 23 02:57:21 PDT 1997 */
//	printf("sec = %d\n", sec);
#endif /* 0 - Tristan Savatier - Sat Aug 23 02:57:21 PDT 1997 */
	
	if (sec >= 0) {
		int		hh, mm, ss;

		hh = sec / 3600;
		mm = (sec / 60) % 60;
		ss = sec % 60;
		
		if (hh > 99) {
			/*
			 * This can only happen if an absurd value of Video_bitrate
			 * or Mux_bitrate is received.  But if we don't test that,
			 * some bad streams can (and do) cause mtv to crash.
			 */
			sprintf(buffer, "??:??:??");
		} else {
			sprintf(buffer, "%02d:%02d:%02d", hh, mm, ss);
		}
		if (buffer[sizeof buffer - 1] != 0) abort_now();
		
		return buffer;
	} else {
		return "n/a";
	}
}


/*--------------------------------------------------------------------------+
| copy_str																	|
+--------------------------------------------------------------------------*/
static void
copy_str(dst, src, size)
	char *dst;
	const char *src;
	u_int	size;
{
	strncpy(dst, src, size - 1);
	dst[size - 1] = 0;
}


/*--------------------------------------------------------------------------+
| get_position_parameters													|
+--------------------------------------------------------------------------*/
static void
get_position_parameters()
{
	G2A_SEND_VOID(COM_GET_POS, 0);
	if (!Seeking) {
		G2A_SEND_VOID(COM_GET_NUMBER_OF_FRAMES_PLAYED, 0);
		if (Gui_state == STATE_PLAYING) {
			G2A_SEND_VOID(COM_GET_ACTUAL_FRAMERATE, 0);
		}
	}
}

/*--------------------------------------------------------------------------+
| my_idle																	|
+--------------------------------------------------------------------------*/
static int
my_idle(xev, data)
	const XEvent *xev;
	const void *data;
{
	double time_now;
	static double Last_time;
	
	if (Seeking) {
		return 0;
	}
	if (Gui_state != STATE_PLAYING && !Fast_forward) {
		return 0;
	}
	
	time_now = get_time_now();
	if (time_now - Last_time > SLIDER_UPDATE_FREQUENCY) {
		Last_time = time_now;
		get_position_parameters();
	}
	return 0;
}

/*--------------------------------------------------------------------------+
| at_close_free																|
+--------------------------------------------------------------------------*/
static int 
at_close_free(fl_form, data)
	FL_FORM *fl_form;
	void	*data;
{
	fl_hide_form(fl_form);
	fl_free_form(fl_form);
	return FL_IGNORE;
}

/*--------------------------------------------------------------------------+
| at_close_hide																|
+--------------------------------------------------------------------------*/
static int 
at_close_hide(fl_form, data)
	FL_FORM *fl_form;
	void	*data;
{
	fl_hide_form(fl_form);
	return FL_IGNORE;
}

/*--------------------------------------------------------------------------+
| file_browser_at_close_hide												|
+--------------------------------------------------------------------------*/
static int 
file_browser_at_close_hide(fl_form, data)
	FL_FORM *fl_form;
	void	*data;
{
	fl_hide_form(fl_form);
	
#if FL_REVISION == 81
	/*
	 * A bug of some sort. File selector disables all
	 * other forms, so when it goes away this way, 
	 * all forms remain disabled. This will be fixed.
	 * For the time being,  you can add fl_activate_all_forms
	 * after fl_hide_form.
	 */
	fl_activate_all_forms();
#endif

	return FL_IGNORE;
}

/*--------------------------------------------------------------------------+
| at_close_controls															|
+--------------------------------------------------------------------------*/
static int 
at_close_controls(fl_form, data)
	FL_FORM *fl_form;
	void	*data;
{
	command(NULL, COM_QUIT);
	return FL_OK;
}

/*--------------------------------------------------------------------------+
| show_and_raise_form														|
+--------------------------------------------------------------------------*/
static void
show_and_raise_form(form, place, border, title)
	FL_FORM *form;
	int		place;
	int		border;
	const char *title;
{
	fl_show_form(form, place, border, title);
	fl_raise_form(form);
}

/*--------------------------------------------------------------------------+
| show_alert																|
+--------------------------------------------------------------------------*/
void
show_alert(s1, s2, s3, n)
	const char *s1;
	const char *s2;
	const char *s3;
	int		n;	/* not used in this replacement routine... */
{
	FD_alert *alert;
	
	/*
	 * This is a replacement for Xform's fl_show_alert() routine.
	 * There is a bug in Xforms-0.81 that causes text in fl_show_alert() windows
	 * to get garbled when several alerts windows are opened at the same time.
	 * Note: fl_show_alert() does not work well in Xforms-0.86 either:
	 * Alert windows disapear when fl_show_alert() is called several times
	 * successively (without dismissing the previous alert window), which can
	 * happen. We like better our custom non-modal alerts.
	 * This replacement routine prevents all the problems...
	 */

	if ((alert = create_form_alert()) == NULL) {
		FATAL_ERROR(("cannot create Alert form"));
	}
	fl_set_object_label(alert->s1, s1);
	fl_set_object_label(alert->s2, s2);
	fl_set_object_label(alert->s3, s3);
	
	fl_set_form_atclose(alert->alert, at_close_free, NULL);
	fl_set_form_icon(alert->alert, Icon_pixmap, 0);

	show_and_raise_form(alert->alert, FL_PLACE_CENTER, FL_FULLBORDER, "MpegTV Alert");
}

/*--------------------------------------------------------------------------+
| show_alert_message														|
+--------------------------------------------------------------------------*/
void
show_alert_message(msg1, msg2)
	const char *msg1;
	const char *msg2;
{
	const char *s;
	
	if (errno != 0) {
		s = strerror(errno);
	} else {
		s = NULL;
	}

	show_alert(msg1, msg2, s == NULL ? "" : s, 1);
}

/*--------------------------------------------------------------------------+
| reset_loop																|
+--------------------------------------------------------------------------*/
static void
reset_loop()
{
	/* reset Auto_seek to 0 */
	if (User_options.loop_start_pos != 0) {
		G2A_SEND_INT(COM_LOOP_START_POS, 0, 0);
		User_options.loop_start_pos = 0;
	}
	if (User_options.loop_end_pos != 0) {
		G2A_SEND_INT(COM_LOOP_END_POS, 0, 0);
		User_options.loop_end_pos = 0;
	}
}

/*--------------------------------------------------------------------------+
| reset_auto_seek															|
+--------------------------------------------------------------------------*/
static void
reset_auto_seek()
{
	/* reset Auto_seek to 0 */
	G2A_SEND_INT(COM_AUTO_SEEK, 0, 0);
}

/*--------------------------------------------------------------------------+
| reset_loop_if_pos_out_of_loop												|
+--------------------------------------------------------------------------*/
static void
reset_loop_if_pos_out_of_loop(n)
	off_t	n;
{
	/* reset Auto_seek and loop_end_pos 0 is n is out of the loop */
	
	if ((User_options.loop_start_pos != 0 && n < User_options.loop_start_pos) || (User_options.loop_end_pos != 0 && n > User_options.loop_end_pos)) {
		G2A_SEND_INT(COM_LOOP_START_POS, 0, 0);
		G2A_SEND_INT(COM_LOOP_END_POS, 0, 0);
		User_options.loop_start_pos = 0;
		User_options.loop_end_pos = 0;
	}
}

/*--------------------------------------------------------------------------+
| file_command																|
+--------------------------------------------------------------------------*/
static void
file_command(p, sel_number, message, com)
	SEL_STATE	*p;
	int		sel_number;
	const char *message;
	COMMAND	com;
{
	const char *answer;

	/*
	 * This code makes the file selector smarter than
	 * the default fl_show_fselector() routine.
	 */
	
	fl_use_fselector(sel_number);
	
	fl_set_form_atclose(fl_get_fselector_form(), file_browser_at_close_hide, NULL);
	fl_set_form_icon(fl_get_fselector_form(), Icon_pixmap, 0);
	
	fl_set_fselector_title(message);
	fl_invalidate_fselector_cache();

	answer = fl_show_fselector (message,
									p->directory,
									p->pattern,
									p->filename);

	/* remember state for next time! */
	copy_str(p->pattern, (const char *)fl_get_pattern(), sizeof p->pattern);
	copy_str(p->directory, (const char *)fl_get_directory(), sizeof p->directory);
	copy_str(p->filename, (const char *)fl_get_filename(), sizeof p->filename);
	
	if (answer == NULL) {
		return;
	}
	if (answer[0] == 0) {
		errno = 0;
		show_alert_message("No file selected!", "");
		return;
	}
	G2A_SEND_STRING(com, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, answer);
}

/*--------------------------------------------------------------------------+
| select_file_in_directory													|
+--------------------------------------------------------------------------*/
static void
select_file_in_directory(dirname)
	const char *dirname;
{
	SEL_STATE	*p = &Open_file;
	
	if (dirname[0] != 0) {
		if (chdir(dirname) != 0) {
			FATAL_ERROR(("cannot chdir to %s", dirname));
		}
		copy_str(p->directory, (const char *)dirname, sizeof p->directory);
	}

	file_command(&Open_file, 0, "Open MPEG file:", COM_OPEN_FILE);
}

/*--------------------------------------------------------------------------+
| file_open																	|
+--------------------------------------------------------------------------*/
static void
file_open()
{
	reset_auto_seek();
	reset_loop();
	file_command(&Open_file, 0, "Open MPEG file:", COM_OPEN_FILE);
}


/*--------------------------------------------------------------------------+
| file_save																	|
+--------------------------------------------------------------------------*/
static void
file_save()
{
	file_command(&Save_file, 1, "Save MPEG file as:", COM_SAVE_FILE);
}

/*--------------------------------------------------------------------------+
| image_save_ppm															|
+--------------------------------------------------------------------------*/
static void
image_save_ppm()
{
	file_command(&Save_image_ppm, 2, "Save Image in PPM file:", COM_SAVE_IMAGE_PPM);
}


/*--------------------------------------------------------------------------+
| image_save_tga															|
+--------------------------------------------------------------------------*/
static void
image_save_tga()
{
	file_command(&Save_image_tga, 3, "Save Image in TGA file:", COM_SAVE_IMAGE_TGA);
}


/*--------------------------------------------------------------------------+
| hide_seek_controls														|
+--------------------------------------------------------------------------*/
static void 
hide_seek_controls()
{
	fl_hide_object(Controls->frame_slider);
#if 0 /* Tristan Savatier - Sun Aug 3 02:32:25 PDT 1997 */
//	fl_hide_object(Controls->time_elapsed);
#endif /* 0 - Tristan Savatier - Sun Aug 3 02:32:25 PDT 1997 */
	fl_hide_object(Controls->time_total);
	fl_hide_object(Controls->rewind);
	fl_hide_object(Controls->end);
	fl_hide_object(Controls->loop);

	fl_set_slider_value(Controls->frame_slider, 0.0);
	
	fl_set_menu_item_mode(Controls->file_menu, FILE_MENU_SAVE_AS, FL_PUP_GREY);
}

/*--------------------------------------------------------------------------+
| show_seek_controls														|
+--------------------------------------------------------------------------*/
static void 
show_seek_controls()
{
	fl_show_object(Controls->frame_slider);
	fl_show_object(Controls->time_elapsed);
	fl_show_object(Controls->time_total);
	fl_show_object(Controls->rewind);
	fl_show_object(Controls->end);
	fl_show_object(Controls->loop);

	fl_set_menu_item_mode(Controls->file_menu, FILE_MENU_SAVE_AS, FL_PUP_NONE);
}

/*--------------------------------------------------------------------------+
| disable_audio_infos														|
+--------------------------------------------------------------------------*/
static void 
disable_audio_infos()
{
	fl_set_object_label(Infos->mux_bitrate, "n/a");
	fl_set_object_label(Infos->audio_bitrate, "n/a");
	fl_set_object_label(Infos->audio_mode, "n/a");
	fl_set_object_label(Infos->audio_layer, "n/a");
	fl_set_object_label(Infos->audio_frequency, "n/a");
	fl_set_object_label(Infos->audio_copyright, "n/a");
	fl_set_object_label(Infos->audio_original, "n/a");
	fl_set_object_label(Infos->audio_protected, "n/a");
}

/*--------------------------------------------------------------------------+
| disable_stream_infos														|
+--------------------------------------------------------------------------*/
static void 
disable_stream_infos()
{
	Bits_filesize = -1;
	Video_bitrate = 0;
	Mux_bitrate = 0;
	
	fl_set_object_label(Infos->file_name, "n/a");
	fl_set_object_label(Infos->file_size, "n/a");
	fl_set_object_label(Infos->video_width, "n/a");
	fl_set_object_label(Infos->video_height, "n/a");
	fl_set_object_label(Infos->video_bitrate, "n/a");
	fl_set_object_label(Infos->video_framerate, "n/a");
	fl_set_object_label(Infos->video_vbv_buffer_size, "n/a");
	fl_set_object_label(Infos->video_constrained, "n/a");
	fl_set_object_label(Infos->video_compliant, "n/a");
	disable_audio_infos();
}

/*--------------------------------------------------------------------------+
| hide_audio_volume_and_infos												|
+--------------------------------------------------------------------------*/
static void 
hide_audio_volume_and_infos()
{
	fl_hide_object(Controls->audio_volume_slider);
	fl_hide_object(Controls->audio_balance_slider);
}

/*--------------------------------------------------------------------------+
| show_audio_volume															|
+--------------------------------------------------------------------------*/
static void 
show_audio_volume()
{
	fl_show_object(Controls->audio_volume_slider);
}

/*--------------------------------------------------------------------------+
| show_audio_balance														|
+--------------------------------------------------------------------------*/
static void 
show_audio_balance()
{
	fl_show_object(Controls->audio_balance_slider);
}

/*--------------------------------------------------------------------------+
| hide_audio_controls														|
+--------------------------------------------------------------------------*/
static void 
hide_audio_controls()
{
	fl_hide_object(Controls->audio_mute);

	hide_audio_volume_and_infos();
}

/*--------------------------------------------------------------------------+
| show_audio_controls														|
+--------------------------------------------------------------------------*/
static void 
show_audio_controls()
{
	fl_show_object(Controls->audio_mute);

	show_audio_volume();
	show_audio_balance();
}



/*--------------------------------------------------------------------------+
| hide_stream_controls														|
+--------------------------------------------------------------------------*/
static void 
hide_stream_controls()
{
	fl_hide_object(Controls->actual_framerate);
	fl_hide_object(Controls->actual_framerate_label);
	fl_hide_object(Controls->play_pause);
	fl_hide_object(Controls->stop);
	fl_hide_object(Controls->step);
	fl_hide_object(Controls->fast_forward);
	fl_hide_object(Controls->zoom);
	fl_hide_object(Controls->number_of_frames_played);
	fl_hide_object(Controls->number_of_frames_played_label);

	fl_hide_object(Controls->time_elapsed);
	
	fl_set_menu_item_mode(Controls->file_menu, FILE_MENU_INFOS, FL_PUP_GREY);
	fl_set_menu_item_mode(Controls->file_menu, FILE_MENU_CLOSE, FL_PUP_GREY);
	fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_RAISE, FL_PUP_GREY);
	fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_FRAME_INFOS, FL_PUP_GREY);
	fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_SAVE_PPM, FL_PUP_GREY);
	fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_SAVE_TGA, FL_PUP_GREY);

	hide_seek_controls();
	hide_audio_controls();
	disable_stream_infos();
}

/*--------------------------------------------------------------------------+
| show_stream_controls														|
+--------------------------------------------------------------------------*/
static void 
show_stream_controls()
{
	fl_show_object(Controls->actual_framerate);
	fl_show_object(Controls->actual_framerate_label);
	fl_show_object(Controls->play_pause);
	fl_show_object(Controls->stop);
	fl_show_object(Controls->step);
	if (fl_get_button(Sync->sync_sync) || fl_get_button(Sync->sync_nominal)) {
		fl_show_object(Controls->fast_forward);
	}
	fl_show_object(Controls->zoom);
	fl_show_object(Controls->number_of_frames_played);
	fl_show_object(Controls->number_of_frames_played_label);

	fl_show_object(Controls->time_elapsed);
	
	fl_set_menu_item_mode(Controls->file_menu, FILE_MENU_INFOS, FL_PUP_NONE);
	fl_set_menu_item_mode(Controls->file_menu, FILE_MENU_CLOSE, FL_PUP_NONE);
	fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_RAISE, FL_PUP_NONE);
	fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_FRAME_INFOS, FL_PUP_NONE);
	fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_SAVE_PPM, FL_PUP_NONE);
	fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_SAVE_TGA, FL_PUP_NONE);
}

/*--------------------------------------------------------------------------+
| u2str																		|
+--------------------------------------------------------------------------*/
static const char *
u2str(n)
	u_int	n;
{
	static char buffer[11];
	
	/* buffer[sizeof buffer - 1] = 0; */
	sprintf(buffer, "%u", n);
	if (buffer[sizeof buffer - 1] != 0) abort_now();
	
	return buffer;
}

/*--------------------------------------------------------------------------+
| d2str																		|
+--------------------------------------------------------------------------*/
static const char *
d2str(d, decimals)
	double	d;
	int		decimals;
{
	static char buffer[10];

	/* buffer[sizeof buffer - 1] = 0; */
	sprintf(buffer, "%1.*f", decimals, d);
	if (buffer[sizeof buffer - 1] != 0) abort_now();

	return buffer;
}

/*--------------------------------------------------------------------------+
| clear_dither_buttons														|
+--------------------------------------------------------------------------*/
static void
clear_dither_buttons()
{
	fl_set_button(Video->video_dither_none, 0);
	fl_set_button(Video->video_dither_gray, 0);
	fl_set_button(Video->video_dither_ordered, 0);
	fl_set_button(Video->video_dither_color, 0);
}

/*--------------------------------------------------------------------------+
| set_dither_type_controls													|
+--------------------------------------------------------------------------*/
static void
set_dither_type_controls(dither_type)
	DITHER_TYPE		dither_type;
{
	clear_dither_buttons();

	switch (dither_type) {
		case DITHER_NONE:
			fl_set_button(Video->video_dither_none, 1);
			break;
		case DITHER_GRAY:
			fl_set_button(Video->video_dither_gray, 1);
			break;
		case DITHER_ORDERED:
			fl_set_button(Video->video_dither_ordered, 1);
			break;
		case DITHER_COLOR:
			fl_set_button(Video->video_dither_color, 1);
			break;
		default:
			abort_now();
	}
}

/*--------------------------------------------------------------------------+
| set_sync_type_controls													|
+--------------------------------------------------------------------------*/
static void
set_sync_type_controls(sync_type)
	SYNC_TYPE sync_type;
{
	FL_OBJECT	*obj;
	
	switch (sync_type) {
		case SYNC_FRAMERATE:	obj = Sync->sync_nominal; break;
		case SYNC_I:			obj = Sync->sync_i; break;
		case SYNC_IP:			obj = Sync->sync_ip; break;
		case SYNC_IPB:			obj = Sync->sync_ipb; break;
		default:	abort_now();
	}
	fl_set_button(obj, 1);
}

/*--------------------------------------------------------------------------+
| set_file_selector_pattern													|
+--------------------------------------------------------------------------*/
static void
set_file_selector_pattern(filename)
	const char *filename;
{
	const char *name;	/* last part of file pathname */
	const char *extension;	/* file extension, if any */
	const char *p;
	char    pattern[PATH_MAX];

	assert(filename != NULL);
	
	if ((p = rindex(filename, '/')) == NULL) {
		name = filename;
	} else {
		name = p + 1;
	}
	
	pattern[sizeof pattern - 1] = '\0';
	
	if ((extension = rindex(name, '.')) != NULL) {
		sprintf (pattern, "*%s", extension);

		if (pattern[sizeof pattern - 1] != '\0') {
			/* buffer overflow! */
			abort_now();
		}
	
		copy_str(Open_file.pattern, pattern, sizeof Open_file.pattern);
	}
}

/*--------------------------------------------------------------------------+
| set_stream_file_info														|
+--------------------------------------------------------------------------*/
static void
set_stream_file_info(filename)
	const char *filename;
{
	char dir[PATH_MAX];
	const char *name;	/* last part of file pathname */
	const char *p;
	char    window_title[PATH_MAX];

	assert(filename != NULL);
	
	window_title[sizeof window_title - 1] = '\0';

	if ((p = rindex(filename, '/')) == NULL) {
		strcpy(dir, ".");
		name = filename;
		sprintf (window_title, " MpegTV - %s, dir: .", filename);
	} else {
		strncpy(dir, filename, (size_t)(p - filename));
		dir[p - filename] = 0;
		name = p + 1;
		sprintf (window_title, " MpegTV - %s, dir: %.*s", name, name - filename, filename);
	}
	
	if (window_title[sizeof window_title - 1] != '\0') {
		/* buffer overflow! */
		abort_now();
	}

	/* set the Player's window and icon name */
	G2A_SEND_STRING(COM_WINDOW_NAME, 0, window_title);
	G2A_SEND_STRING(COM_ICON_NAME, 0, name);
	
	fl_set_object_label(Infos->file_name, name);

	copy_str(Open_file.filename, name, sizeof Open_file.filename);
	copy_str(Open_file.directory, dir, sizeof Open_file.directory);
}

#if 0 /* Tristan Savatier - Thu Aug 28 20:58:46 PDT 1997 */
///*--------------------------------------------------------------------------+
//| set_framerate																|
//+--------------------------------------------------------------------------*/
//static void
//set_framerate()
//{
//	Specified_framerate_is_nominal = False;
//	if (fl_get_button(Sync->sync_sync)) {
//		G2A_SEND_DOUBLE(COM_SYNC_FRAMERATE_SET, 0, (double)fl_get_slider_value(Sync->sync_framerate_slider));
//		Framerate_is_nominal = False;
//	}
//}
#endif /* 0 - Tristan Savatier - Thu Aug 28 20:58:46 PDT 1997 */

/*--------------------------------------------------------------------------+
| set_sync_sync																|
+--------------------------------------------------------------------------*/
static void
set_sync_sync()
{
	/* fast forward enabled */
	if (Controls->play_pause->visible) {
		fl_show_object(Controls->fast_forward);
	}
	G2A_SEND_INT(COM_SYNC_TYPE, 0, SYNC_FRAMERATE);
	if (Specified_framerate_is_nominal) {
		G2A_SEND_DOUBLE(COM_SYNC_FRAMERATE_SET, 0, Nominal_framerate);
		Framerate_is_nominal = True;
	} else {
		G2A_SEND_DOUBLE(COM_SYNC_FRAMERATE_SET, 0, (double)fl_get_slider_value(Sync->sync_framerate_slider));
		Framerate_is_nominal = False;
	}
}

/*--------------------------------------------------------------------------+
| gui_set_user_options														|
+--------------------------------------------------------------------------*/
static void
gui_set_user_options()
{
	/*
	 * Ask to receive ButtonRelease events in Player's window.
	 * On some systems (Solaris, IRIX) it seems to be necessary
	 * to ask for ButtonPress events in order to get the ButtonRelease
	 * events. No idea why!
	 */
	G2A_SEND_INT(COM_X_EVENT_MASK, 0, (int)(ButtonPressMask|ButtonReleaseMask));
#if 0 /* Tristan Savatier - Tue Aug 12 01:24:24 PDT 1997 */
//	G2A_SEND_INT(COM_X_EVENT_MASK, 0, (int)(ButtonReleaseMask));
#endif /* 0 - Tristan Savatier - Tue Aug 12 01:24:24 PDT 1997 */

	if (User_options.audio_mode == AUDIO_MODE_NONE) {
		fl_set_menu_item_mode(Controls->options_menu, OPTIONS_MENU_AUDIO, FL_PUP_GREY);
	}

	G2A_SEND_INT(COM_LOOP, 0, (int)User_options.loop);
	fl_set_button(Controls->loop, User_options.loop);

	G2A_SEND_INT(COM_ZOOM, 0, (int)User_options.zoom);
	fl_set_button(Controls->zoom, (int)User_options.zoom);
	
	G2A_SEND_INT(COM_AUTO_EXIT, 0, (int)User_options.auto_exit);
	/* fl_set_button(Controls->auto_exit, User_options.auto_exit); */

	G2A_SEND_INT(COM_AUTO_PLAY, 0, (int)User_options.auto_play);
	fl_set_button(Controls->auto_play, User_options.auto_play);

	if (User_options.dither_type != (DITHER_TYPE)-1) {
		G2A_SEND_INT(COM_VIDEO_DITHER_TYPE, 0, (int)User_options.dither_type);
	}
	
	if (User_options.framerate == 0) {
		G2A_SEND_INT(COM_SYNC_TYPE, 0, (int)User_options.sync_type);
		set_sync_type_controls((SYNC_TYPE)User_options.sync_type);
	} else {
#if 0 /* Tristan Savatier - Thu Aug 28 20:52:38 PDT 1997 */
//		fl_set_button(Sync->sync_sync, 1);
//		G2A_SEND_INT(COM_SYNC_TYPE, 0, SYNC_FRAMERATE);
#endif /* 0 - Tristan Savatier - Thu Aug 28 20:52:38 PDT 1997 */

		fl_set_slider_value(Sync->sync_framerate_slider, User_options.framerate);
		fl_set_button(Sync->sync_sync, 1);
		Specified_framerate_is_nominal = False;
		set_sync_sync();
#if 0 /* Tristan Savatier - Thu Aug 28 20:48:23 PDT 1997 */
//		Specified_framerate_is_nominal = False;
//		G2A_SEND_DOUBLE(COM_SYNC_FRAMERATE_SET, 0, User_options.framerate);
//		Framerate_is_nominal = False;
#endif /* 0 - Tristan Savatier - Thu Aug 28 20:48:23 PDT 1997 */
	}

	/*
	 * We dont need to use COM_MSG_SELECT anymore because
	 * the message selector mask
	 * is now passed through option -w when mtvp is started.
	 */
	/* G2A_SEND_INT(COM_MSG_SELECT, 0, (int)User_options.msg_select); */
	
	fl_set_slider_value(Video->video_gamma_slider, User_options.gamma);
	command(Video->video_gamma_slider, COM_VIDEO_GAMMA_SLIDER);
	
	/*
	 * We dont need to use COM_AUDIO_MSG_SELECT anymore because
	 * the audio message selector mask
	 * is now passed through option -aw when mtvp is started.
	 */
	/* G2A_SEND_INT(COM_AUDIO_MSG_SELECT, 0, (int)User_options.audio_msg_select); */

	G2A_SEND_INT(COM_AUDIO_QUALITY, 0, (int)User_options.audio_quality);
	switch (User_options.audio_quality) {
		case AUDIO_QUALITY_LOW: fl_set_button(Audio->audio_quality_low, 1); break;
		case AUDIO_QUALITY_MEDIUM: fl_set_button(Audio->audio_quality_medium, 1); break;
		case AUDIO_QUALITY_HIGH: fl_set_button(Audio->audio_quality_high, 1); break;
		default:	abort_now();
	}
	
	G2A_SEND_INT(COM_AUDIO_CHANNEL, 0, (int)User_options.audio_channel);
	switch (User_options.audio_channel) {
		case AUDIO_CHANNEL_LEFT: fl_set_button(Audio->audio_ch_left, 1); break;
		case AUDIO_CHANNEL_RIGHT: fl_set_button(Audio->audio_ch_right, 1); break;
		case AUDIO_CHANNEL_BOTH: fl_set_button(Audio->audio_ch_both, 1); break;
		default:	abort_now();
	}

	G2A_SEND_INT(COM_AUDIO_CONTROLS, 0, AUDIO_CONTROLS_PCM);

	if (User_options.audio_pcm_volume >= 0) {
		G2A_SEND_DOUBLE(COM_AUDIO_VOLUME, 0, User_options.audio_pcm_volume);
	}
	if (User_options.audio_pcm_balance >= 0) {
		G2A_SEND_DOUBLE(COM_AUDIO_BALANCE, 0, User_options.audio_pcm_balance);
	}
	
	G2A_SEND_INT(COM_AUDIO_CONTROLS, 0, AUDIO_CONTROLS_MASTER);

	if (User_options.audio_master_volume >= 0) {
		G2A_SEND_DOUBLE(COM_AUDIO_VOLUME, 0, User_options.audio_master_volume);
	}
	if (User_options.audio_master_balance >= 0) {
		G2A_SEND_DOUBLE(COM_AUDIO_BALANCE, 0, User_options.audio_master_balance);
	}

	switch (User_options.audio_controls) {
		case AUDIO_CONTROLS_MASTER:
			G2A_SEND_INT(COM_AUDIO_CONTROLS, 0, AUDIO_CONTROLS_MASTER);
			fl_set_button(Audio->audio_controls_master, 1);
			break;
		case AUDIO_CONTROLS_PCM:
			G2A_SEND_INT(COM_AUDIO_CONTROLS, 0, AUDIO_CONTROLS_PCM);
			fl_set_button(Audio->audio_controls_pcm, 1);
			break;
		default:	abort_now();			
	}
	if (User_options.audio_ports != -1) {
		G2A_SEND_INT(COM_AUDIO_PORTS_SELECT, 0, User_options.audio_ports);
	}


	if (User_options.loop_start_pos != 0) {
		G2A_SEND_INT(COM_AUTO_SEEK, 0, (int)(u_int)User_options.loop_start_pos);
		G2A_SEND_INT(COM_LOOP_START_POS, 0, (int)(u_int)User_options.loop_start_pos);
	}
	if (User_options.loop_end_pos != 0) {
		G2A_SEND_INT(COM_LOOP_END_POS, 0, (int)(u_int)User_options.loop_end_pos);
	}

	G2A_SEND_INT(COM_WINDOW_POSITION_X, 0, User_options.mtvp_geometry_x);
	G2A_SEND_INT(COM_WINDOW_POSITION_Y, 0, User_options.mtvp_geometry_y);

	if (User_options.stdin_mode) {
		G2A_SEND_INT(COM_OPEN_FD, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, 0);
	} else if (User_options.port != 0) {
		G2A_SEND_INT(COM_OPEN_PORT, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, User_options.port);
	} else if (User_options.port_udp != 0) {
		G2A_SEND_INT(COM_OPEN_PORT_UDP, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, User_options.port_udp);
	} else if (User_options.filename != NULL) {
		int		fd;
		struct stat stat_buf;
	
		fd = open(User_options.filename, O_RDONLY|O_BINARY);
		if (fd == -1) {
			/* FATAL_ERROR(("Error opening input file: %s", User_options.filename)); */
			show_alert("Error opening input file:", User_options.filename, strerror(errno), 1);
		} else {
			if (fstat(fd, &stat_buf) != 0) {
				show_alert("Error accessing input file:", User_options.filename, strerror(errno), 1);
			} else {
		
				if (S_ISDIR(stat_buf.st_mode)) {
		
					select_file_in_directory(User_options.filename);
				} else {
		
					G2A_SEND_STRING(COM_OPEN_FILE, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, User_options.filename);
					set_file_selector_pattern(User_options.filename);
				}
			}
			close(fd);
		}
	}

}

/*--------------------------------------------------------------------------+
| set_time_total															|
+--------------------------------------------------------------------------*/
static void
set_time_total()
{
	const char *time_str;
	int	sec;

	sec = get_elapsed_seconds(Bits_filesize);
	
	time_str = get_time_string(sec);

	fl_set_object_label(Controls->time_total, time_str);
	
	fl_set_object_label(Infos->file_size, Bits_filesize == -1 ? "n/a" : u2str(UDIV_ROUNDED((u_int)Bits_filesize, 1024)));

	if (Bits_filesize != -1) {
		fl_set_slider_bounds(Controls->frame_slider, 0.0, (double)Bits_filesize);
	}
}

/*--------------------------------------------------------------------------+
| set_time_elapsed															|
+--------------------------------------------------------------------------*/
static void
set_time_elapsed(pos)
	int		pos;
{
	const char *time_str;
	int	sec;

	sec = get_elapsed_seconds(pos);
	time_str = get_time_string(sec);
	
#if 0 /* Tristan Savatier - Mon Aug 18 00:42:25 PDT 1997 */
//	printf("set_time_elapsed: pos = %d, sec = %d, time_str = %s\n", pos, sec, time_str);
#endif /* 0 - Tristan Savatier - Mon Aug 18 00:42:25 PDT 1997 */
	
	fl_set_object_label(Controls->time_elapsed, time_str);
}

/*--------------------------------------------------------------------------+
| disable_frame_infos														|
+--------------------------------------------------------------------------*/
static void
disable_frame_infos()
{
	fl_set_object_label(Frame_infos->frame_type, "n/a");
	fl_set_object_label(Frame_infos->frame_temporal_reference, "n/a");
	fl_set_object_label(Frame_infos->frame_forward_f_code, "n/a");
	fl_set_object_label(Frame_infos->frame_forward_full_pel, "n/a");
	fl_set_object_label(Frame_infos->frame_backward_f_code, "n/a");
	fl_set_object_label(Frame_infos->frame_backward_full_pel, "n/a");
	fl_set_object_label(Frame_infos->frame_vbv_delay, "n/a");
}


/*--------------------------------------------------------------------------+
| update_control_title														|
+--------------------------------------------------------------------------*/
static void
update_control_title()
{
	const char *str;
	
	switch (Gui_state) {
	case STATE_STOPPED:	str = "MpegTV Controls (Stopped)"; break;
	case STATE_PAUSED:	str = "MpegTV Controls (Paused)"; break;
	case STATE_PLAYING:	str = "MpegTV Controls (Playing)"; break;
	default:			abort_now();
	}
	fl_set_form_title(Controls->controls, str);
}


/*--------------------------------------------------------------------------+
| update_state																|
+--------------------------------------------------------------------------*/
static void
update_state(state)
	STATE	state;
{
	Gui_state = state;
	switch (state) {
	case STATE_STOPPED:
		get_position_parameters();
		fl_set_object_label(Controls->play_pause, "@>");
		fl_set_object_label(Controls->actual_framerate, "n/a");
		Gui_play_pause_set_state = STATE_PLAYING;
		break;
	case STATE_PAUSED:
		get_position_parameters();
		fl_set_object_label(Controls->play_pause, "@>");
		fl_set_object_label(Controls->actual_framerate, "n/a");
		Gui_play_pause_set_state = STATE_PLAYING;
		break;
	case STATE_PLAYING:
		fl_set_object_label(Controls->play_pause, "||");
		Gui_play_pause_set_state = STATE_PAUSED;
		disable_frame_infos();
		break;
	default:
		abort_now();
	}
	update_control_title();
}

/*--------------------------------------------------------------------------+
| send_seek																	|
+--------------------------------------------------------------------------*/
static void
send_seek(n)
	int		n;
{
	reset_loop_if_pos_out_of_loop(n);
	G2A_SEND_INT(COM_SEEK, 0, n);
	Not_ready = True;
}

/*--------------------------------------------------------------------------+
| process_alert																|
+--------------------------------------------------------------------------*/
static void
process_alert(data)
	PMP_DATA_ALERT	*data;
{
	const char *s1;
	const char *s2;
	const char *s3;

	s1 = "Warning!";
	
	if (data->error != 0) {
		s3 = strerror(data->error);
	} else {
		s3 = "";
	}
	switch (data->alert) {
		case ALERT_BAD_STREAM:
			hide_stream_controls();
			s2 = "Hum... This doesn't look like an MPEG Systems or Video stream";
			s3 = "Note: MpegTV Player 1.0 cannot play MPEG Audio files";
			break;
		case ALERT_NO_AUDIO:
			s2 = "Audio device busy or sampl. rate unsup. (muting)";
			break;
		/* Note: ALERT_DEMO has become obsolete */
		/* Note: ALERT_DEMO_MUTE has become obsolete */
		case ALERT_NO_SHARED_MEMORY:
			s2 = "Can't share video memory with X server: performances will be reduced!";
			break;
		case ALERT_USING_PRIVATE_COLORMAP:
			fl_set_button(Video->video_dither_ordered_colormap_private, 1);
			s2 = "Not enough colors available in shared colormap";
			s3 = "Using private colormap!";
			break;
		case ALERT_MPEG2_VIDEO:
			hide_stream_controls();
			s2 = "Hum... This looks like an MPEG-2 stream";
			s3 = "MpegTV Player 1.0 only plays MPEG-1 Systems or Video streams";
			break;
		default:
			s2 = "Unknown alert";
			break;
	}

	show_alert(s1, s2, s3, 1);
}


/*--------------------------------------------------------------------------+
| send_state																|
+--------------------------------------------------------------------------*/
static void
send_state(state)
	STATE	state;
{
	G2A_SEND_INT(COM_STATE, 0, (int)state);
}

/*--------------------------------------------------------------------------+
| set_state																	|
+--------------------------------------------------------------------------*/
static void
set_state(state)
	STATE	state;
{
	send_state(state);
	update_state(state);
}



/*--------------------------------------------------------------------------+
| a2g_read_command															|
+--------------------------------------------------------------------------*/
static int
a2g_read_command(fd)
	int		fd;	/* not used, because we use library wrappers like pmp_receive_msg() to read messages */
{
	PMP_HEADER	msg;
	PMP_DATA_A2G	data;
	
	if (a2g_read_msg(&Player, &msg, &data) == -1) {
		FATAL_ERROR(("a2g_read_msg: error reading msg"));
	}

	MSG(MSG_PMP, ("received msg: %s", pmp_print_msg(&msg, (const char *)&data)));

	switch ((A2G_COM)msg.type) {
		case A2G_COM_API_VERSION:
			if (data.data_generic.data_int != COM_API_VERSION) {
				FATAL(("API version mismatch, expecting %d, got %d. The mtvp program found in your PATH is not compatible with this frontend!", COM_API_VERSION, data.data_generic.data_int)); 
			}
			break;
		case A2G_COM_READY:
			/*
			 * Any initial command here will be executed by the player with
			 * a smaller delay, since the player is in Ready state.
			 * The player sends the A2G_COM_READY message when it has completed
			 * all its initializations and is ready to work.
			 */	
			break;
		case A2G_COM_DISPLAY_READY:
			Not_ready = False;
			if (Pending_seek >= 0) {
				send_seek(Pending_seek);
				if (!Seeking) {
					send_state(State_before_seeking);
				}
				G2A_SEND_VOID(COM_GET_POS, 0);
				Pending_seek = -1;
			}
			break;
		case A2G_COM_QUIT:
			EXIT(EXIT_SUCCESS);
			break;
		case A2G_COM_STATE:
			if (!Seeking) {
				update_state((STATE)data.data_generic.data_int);
				if (Gui_state != STATE_PLAYING) {
					get_position_parameters();
				}
			}
			break;
		case A2G_COM_POS:
			if (data.data_generic.data_int != -1) {
				Stream_pos = data.data_generic.data_int;
				
#if 0 /* Tristan Savatier - Sun Aug 3 02:47:25 PDT 1997 */
//				printf("Stream_pos = %d\n", Stream_pos);
#endif /* 0 - Tristan Savatier - Sun Aug 3 02:47:25 PDT 1997 */
				
				set_time_elapsed(Stream_pos);
				if (Bits_filesize != -1 && Stream_pos > Bits_filesize) {
					Bits_filesize = Stream_pos;
					set_time_total();
				}
				if (Pending_seek == -1) {
					fl_set_slider_value(Controls->frame_slider, (double)Stream_pos);
				}
			}
			break;
		case A2G_COM_ACTUAL_FRAMERATE:
			{
				double framerate = data.data_generic.data_double;
				const char *str;
				
				if (framerate == 0) {
					str = "n/a";
				} else {
					str = d2str(framerate, 1);
				}
				fl_set_object_label(Controls->actual_framerate, str);
			}
			break;
		case A2G_COM_NUMBER_OF_FRAMES_PLAYED:
			fl_set_object_label(Controls->number_of_frames_played, u2str((u_int)data.data_generic.data_int));
			break;
		case A2G_COM_AUDIO_ENABLED:
			show_audio_controls();
			break;
		case A2G_COM_AUDIO_DISABLED:
			hide_audio_controls();
			break;
		case A2G_COM_ALERT:
			process_alert(&data.data_alert);
			break;
		case A2G_COM_INPUT_FILE_SIZE:
			Bits_filesize = data.data_generic.data_int;
			if (Bits_filesize == -1) {
				fl_set_button(Controls->loop, 0);
				hide_seek_controls();
			} else {
				show_seek_controls();
			}
			set_time_total();
			fl_set_object_label(Infos->file_size, Bits_filesize == -1 ? "n/a" : u2str(UDIV_ROUNDED((u_int)data.data_generic.data_int, 1024)));
			break;
		case A2G_COM_INPUT_FD:
			break;
		case A2G_COM_INPUT_PORT:
			break;
		case A2G_COM_AUDIO_VOLUME:
			fl_set_slider_value(Controls->audio_volume_slider, data.data_generic.data_double);
			break;
		case A2G_COM_AUDIO_BALANCE:
			fl_set_slider_value(Controls->audio_balance_slider, data.data_generic.data_double);
			break;
		case A2G_COM_AUDIO_MUTE:
			fl_set_button(Controls->audio_mute, 1);
			break;
		case A2G_COM_VIDEO_INFOS:
			{
				const PMP_DATA_VIDEO_INFOS *video_infos = &data.data_video_infos;
				
				fl_set_object_label(Infos->video_width, u2str(video_infos->width));
				fl_set_object_label(Infos->video_height, u2str(video_infos->height));
				
				Video_bitrate = video_infos->bitrate;
				fl_set_object_label(Infos->video_bitrate, Video_bitrate == 0 ? "n/a" : d2str(Video_bitrate*(1.0/1000000), 6));
				set_time_total();

				if (Specified_framerate_is_nominal && Nominal_framerate != video_infos->framerate) {
					Specified_framerate_is_nominal = False;
				}
				Nominal_framerate = video_infos->framerate;
				fl_set_object_label(Infos->video_framerate, d2str(Nominal_framerate, 2));

				/*
				 * Currently mtvp resets automatically the sync framerate to
				 * the nominal framerate when a new file is opened (actually,
				 * it does that when the first sequence header is found). To work
				 * around this problem, mtv resets the framerate to the
				 * currect value after the file has been opened.
				 * This is not very good and may not work in all cases,
				 * because mtvp resets the framerate only once it finds
				 * a sequence header. So if there is garbage at the beginning
				 * of the file, we have a race condition!
				 * The only proper fix to this problem would be to add a new
				 * SYNC mode in mtvp, so that mtvp would know whether or not
				 * it should reset the framerate to nominal.
				 */
				if (fl_get_button(Sync->sync_sync)) {
					if (Specified_framerate_is_nominal) {
						G2A_SEND_DOUBLE(COM_SYNC_FRAMERATE_SET, 0, Nominal_framerate);
						Framerate_is_nominal = True;
					} else {
						G2A_SEND_DOUBLE(COM_SYNC_FRAMERATE_SET, 0, (double)fl_get_slider_value(Sync->sync_framerate_slider));
						Framerate_is_nominal = False;
					}
				} else if (fl_get_button(Sync->sync_nominal)) {
					Framerate_is_nominal = True;
				}


				fl_set_object_label(Infos->video_vbv_buffer_size, video_infos->vbv_buffer_size == 0 ? "n/a" : u2str(UDIV_ROUNDED(video_infos->vbv_buffer_size, 1024)));
				fl_set_object_label(Infos->video_constrained, video_infos->constrained_parameter_flag ? "yes" : "no");
				{
					int		mbwidth = (video_infos->width + 15) / 16;
					int		mbheight = (video_infos->height + 15) / 16;
					
					if (video_infos->constrained_parameter_flag
						&& video_infos->width <= CONSTRAINED_MAX_FRAME_WIDTH
						&& video_infos->height <= CONSTRAINED_MAX_FRAME_HEIGHT
						&& mbwidth * mbheight <= CONSTRAINED_MAX_MB_COUNT
						&& mbwidth * mbheight * video_infos->framerate <= CONSTRAINED_MAX_MB_RATE
						&& video_infos->framerate <= CONSTRAINED_MAX_FRAME_RATE
						&& video_infos->bitrate != 0
						&& video_infos->bitrate <= CONSTRAINED_MAX_BIT_RATE
						&& video_infos->vbv_buffer_size != 0
						&& video_infos->vbv_buffer_size <= CONSTRAINED_MAX_VBV_BUFFER_SIZE) {
						fl_set_object_label(Infos->video_compliant, "maybe");
					} else {
						fl_set_object_label(Infos->video_compliant, "no");
					}
				}
			}
			break;
		case A2G_COM_PACK_INFOS:
			Mux_bitrate = data.data_pack_infos.mux_rate;
			fl_set_object_label(Infos->mux_bitrate, Mux_bitrate == 0 ? "n/a" : d2str(Mux_bitrate*(1.0/1000000), 6));
			set_time_total();
			break;
		case A2G_COM_AUDIO_INFOS:
			{
				const PMP_DATA_AUDIO_INFOS	*audio_infos = &data.data_audio_infos;
				const char *str;
				
				fl_set_object_label(Infos->audio_bitrate, u2str(UDIV_ROUNDED(audio_infos->bitrate, 1000)));
				fl_set_object_label(Infos->audio_layer, u2str((~(audio_infos->layer - 1)) & 3));
				
				switch (audio_infos->mode) {
					case MAL_MODE_STEREO:			str = "stereo"; break;
					case MAL_MODE_JOINT_STEREO:		str = "joint stereo"; break;
					case MAL_MODE_DUAL_CHANNEL:		str = "dual channel"; break;
					case MAL_MODE_SINGLE_CHANNEL:	str = "single channel"; break;
					default: 						abort_now();
				}
				fl_set_object_label(Infos->audio_mode, str);

				fl_set_object_label(Infos->audio_frequency, d2str(audio_infos->sampling_frequency/1000.0, 1));
				fl_set_object_label(Infos->audio_copyright, audio_infos->copyright ? "yes" : "no");
				fl_set_object_label(Infos->audio_original, audio_infos->original ? "original" : "copy");
				fl_set_object_label(Infos->audio_protected, audio_infos->protected_flag ? "yes" : "no");
			}
			break;

		case A2G_COM_VIDEO_DITHER_ENABLE:
			switch (data.data_generic.data_int) {
			case DITHER_NONE:
				fl_show_object(Video->video_dither_none);
				break;
			case DITHER_GRAY:
				fl_show_object(Video->video_dither_gray);
				break;
			case DITHER_ORDERED:
				fl_show_object(Video->video_dither_ordered);
				fl_show_object(Video->video_dither_ordered_colormap_private);
				break;
			case DITHER_COLOR:
				fl_show_object(Video->video_dither_color);
				break;
			default:	abort_now();
			}
			break;
		case A2G_COM_VIDEO_DITHER_TYPE:
			set_dither_type_controls((DITHER_TYPE)data.data_generic.data_int);
			break;
		case A2G_COM_VIDEO_COLORS_ALLOCATED:
			/* printf("%d colors allocated\n", data.data_generic.data_int); */
			break;
		case A2G_COM_FRAME_INFOS:
			{
				const PMP_DATA_FRAME_INFOS *frame_infos = &data.data_frame_infos;
				const char *str;

				switch (frame_infos->mode) {
				case MODE_INTRA:
					str = "I";
					fl_set_object_label(Frame_infos->frame_forward_f_code, "n/a");
					fl_set_object_label(Frame_infos->frame_forward_full_pel, "n/a");
					fl_set_object_label(Frame_infos->frame_backward_f_code, "n/a");
					fl_set_object_label(Frame_infos->frame_backward_full_pel, "n/a");
					break;
				case MODE_PREDICTIVE:
					str = "P";
					fl_set_object_label(Frame_infos->frame_forward_f_code, u2str(frame_infos->f_code[F][HT]));
					fl_set_object_label(Frame_infos->frame_forward_full_pel, frame_infos->full_pel[F] ? "full-pel" : "half-pel");
					fl_set_object_label(Frame_infos->frame_backward_f_code, "n/a");
					fl_set_object_label(Frame_infos->frame_backward_full_pel, "n/a");
					break;
				case MODE_INTERPOLATIVE:
					str = "B";
					fl_set_object_label(Frame_infos->frame_forward_f_code, u2str(frame_infos->f_code[F][HT]));
					fl_set_object_label(Frame_infos->frame_forward_full_pel, frame_infos->full_pel[F] ? "full-pel" : "half-pel");
					fl_set_object_label(Frame_infos->frame_backward_f_code, u2str(frame_infos->f_code[B][HT]));
					fl_set_object_label(Frame_infos->frame_backward_full_pel, frame_infos->full_pel[B] ? "full-pel" : "half-pel");
					break;
				default:					str = "?"; break;
				}
				fl_set_object_label(Frame_infos->frame_type, str);
				fl_set_object_label(Frame_infos->frame_temporal_reference, u2str(frame_infos->temporal_reference));
				fl_set_object_label(Frame_infos->frame_vbv_delay, frame_infos->vbv_delay < 0 ? "vbr" : d2str(frame_infos->vbv_delay, 2));
			}
			break;
		case A2G_COM_GOP_INFOS:
			/* GOP info not yet implemented in mtv! */
#if 0 /* Tristan Savatier - Tue Jun 17 22:49:12 PDT 1997 */
//			{
//				const PMP_DATA_GOP_INFOS *gop_infos = &data.data_gop_infos;
//
//				printf("time_code_drop = %d\n", gop_infos->time_code_drop);
//				printf("time_code_hours = %d\n", gop_infos->time_code_hours);
//				printf("time_code_minutes = %d\n", gop_infos->time_code_minutes);
//				printf("time_code_seconds = %d\n", gop_infos->time_code_seconds);
//				printf("time_code_pictures = %d\n", gop_infos->time_code_pictures);
//				printf("closed_gop = %d\n", gop_infos->closed_gop);
//				printf("broken_link = %d\n", gop_infos->broken_link);
//				printf("\n");
//			}
#endif /* 0 - Tristan Savatier - Tue Jun 17 22:49:12 PDT 1997 */
			break;
		case A2G_COM_ERROR_IN_STREAM:
			/* we ignore those messages */
			break;
		case A2G_COM_AUDIO_PORTS_AVAILABLE:
			Audio_ports = data.data_generic.data_int;
#if 0 /* Tristan Savatier - Sat Aug 16 18:51:12 PDT 1997 */
//			if (Audio_ports == 0) {
//				fl_show_object(Audio->audio_port_line_out);
//			} else
#endif /* 0 - Tristan Savatier - Sat Aug 16 18:51:12 PDT 1997 */
			{
				if (data.data_generic.data_int & AUDIO_PORT_LINE_OUT) {
					fl_show_object(Audio->audio_port_line_out);
				}
				if (data.data_generic.data_int & AUDIO_PORT_HEADPHONE) {
					fl_show_object(Audio->audio_port_headphone);
				}
				if (data.data_generic.data_int & AUDIO_PORT_SPEAKER) {
					fl_show_object(Audio->audio_port_speaker);
				}
			}
			break;
		case A2G_COM_AUDIO_PORTS_SELECTED:
			if (data.data_generic.data_int & AUDIO_PORT_LINE_OUT) {
				fl_set_button(Audio->audio_port_line_out, 1);
			} else {
				fl_set_button(Audio->audio_port_line_out, 0);
			}
			if (data.data_generic.data_int & AUDIO_PORT_HEADPHONE) {
				fl_set_button(Audio->audio_port_headphone, 1);
			} else {
				fl_set_button(Audio->audio_port_headphone, 0);
			}
			if (data.data_generic.data_int & AUDIO_PORT_SPEAKER) {
				fl_set_button(Audio->audio_port_speaker, 1);
			} else {
				fl_set_button(Audio->audio_port_speaker, 0);
			}
			break;
		case A2G_COM_SYSTEMS_END_CODE:
			break;
		case A2G_COM_VIDEO_END_CODE:
			break;
		case A2G_COM_AUDIO_PID:
			/* printf("audio pid = %d\n", data.data_generic.data_int); */
			break;
		case A2G_COM_AUDIO_NO_CONTROLS_SELECT:
			fl_hide_object(Audio->audio_controls_master);
			fl_set_button(Audio->audio_controls_pcm, 1);
			break;
		case A2G_COM_CHALLENGE:
			/*
			 * This is a reserved message of the API.
			 * This private authentification request is sent only if the
			 * player (mtvp) cannot find a valid license key.
			 * The player does not need a valid license key when it
			 * recognizes the bundled mtv front-end binary (precompiled).
			 *
			 * If the player (mtvp) finds a valid license key, it does not
			 * send this message.
			 *
			 * Front-ends (other than the bundled mtv) who receive this message
			 * should advise the user that the player has no license and will
			 * not be functionnal.
			 */
#ifdef CHALLENGE
			handle_challenge(&msg, &data);
#else
			show_alert("This front-end requires a licensed mtvp software", "Use precompiled mtv front-end or purchase", "a development license (see http://www.mpegtv.com for details)", 1);
#endif
			break;
		case A2G_COM_X_EVENT:				
			switch (data.data_x_event.type) {
				case ButtonRelease:
					if (Controls->controls->visible) {
						fl_hide_form(Controls->controls);
					} else {
#if 0 /* Tristan Savatier - Mon Aug 11 03:40:07 PDT 1997 */
//#ifndef LINUX
//							fl_set_form_position(Controls->controls, Controls->controls->x, Controls->controls->y);
//#endif
#endif /* 0 - Tristan Savatier - Mon Aug 11 03:40:07 PDT 1997 */
						show_and_raise_form(Controls->controls, FL_PLACE_POSITION, FL_FULLBORDER, "MpegTV Controls");
						update_control_title();
					}
					break;
				default:
					/* ignore other X events */
					break;
			}
			break;
		case A2G_COM_RETURN_SUCCESS:
			{
				PMP_HEADER	original_msg;
				PMP_DATA_G2A	original_data;
				if (a2g_read_msg(&Player, &original_msg, (PMP_DATA_A2G *)&original_data) == -1) {
					FATAL_ERROR(("a2g_read_msg: error reading msg"));
				}
				MSG(MSG_PMP, ("received msg: %s", pmp_print_msg(&original_msg, (const char *)&original_data)));
				switch (original_msg.type) {
					case COM_OPEN_FILE:
						hide_audio_controls();
						disable_audio_infos();
						disable_stream_infos();
						set_stream_file_info(original_data.data_generic.data_string);
						show_stream_controls();
						break;
					case COM_OPEN_FD:	/*FALLTHROUGH*/
					case COM_OPEN_PORT:	/*FALLTHROUGH*/
					case COM_OPEN_PORT_UDP:
						{
							char type[32];
							char buf[32];
							
							type[sizeof type - 1] = 0;
							if (original_msg.type == COM_OPEN_FD) {
								if (original_data.data_generic.data_int == 0) {
									sprintf(type, "<stdin>");
								} else {
									sprintf(type, "<FD %d>", original_data.data_generic.data_int);
								}
							} else if (original_msg.type == COM_OPEN_PORT) {
								sprintf(type, "<TCP %d>", original_data.data_generic.data_int);
							} else {
								sprintf(type, "<UDP %d>", original_data.data_generic.data_int);
							}
							if (type[sizeof type - 1] != 0) abort_now();
							
							buf[sizeof buf - 1] = 0;
							sprintf(buf, "MpegTV - %s", type);
							if (buf[sizeof buf - 1] != 0) abort_now();
							
							G2A_SEND_STRING(COM_WINDOW_NAME, 0, buf);
							G2A_SEND_STRING(COM_ICON_NAME, 0, type);

							hide_audio_controls();
							disable_audio_infos();
							disable_stream_infos();
							fl_set_object_label(Infos->file_name, type);
							show_stream_controls();
						}
						break;
					case COM_VIDEO_GAMMA:
						if (Gamma.pending != GAMMA_INVALID) {
							G2A_SEND_DOUBLE(COM_VIDEO_GAMMA, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, Gamma.pending);
							Gamma.pending = GAMMA_INVALID;
						} else {							
							Gamma.not_ready = False;
						}
						break;
					case COM_VIDEO_LUMA_FLOOR:
						if (Contrast.pending != CONTRAST_INVALID) {
							G2A_SEND_DOUBLE(COM_VIDEO_LUMA_FLOOR, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, - Contrast.pending);
							Contrast.pending = CONTRAST_INVALID;
						} else {							
							Contrast.not_ready = False;
						}
						break;
					case COM_VIDEO_LUMA_SHIFT:
						if (Brightness.pending != BRIGHTNESS_INVALID) {
							G2A_SEND_DOUBLE(COM_VIDEO_LUMA_SHIFT, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, Brightness.pending);
							Brightness.pending = BRIGHTNESS_INVALID;
						} else {							
							Brightness.not_ready = False;
						}
						break;
					default:
						break;
				}
			}
			break;
		case A2G_COM_RETURN_FAILURE:
			{
				PMP_HEADER	original_msg;
				PMP_DATA_G2A	original_data;
				if (a2g_read_msg(&Player, &original_msg, (PMP_DATA_A2G *)&original_data) == -1) {
					FATAL_ERROR(("a2g_read_msg: error reading msg"));
				}
				MSG(MSG_PMP, ("received msg: %s", pmp_print_msg(&original_msg, (const char *)&original_data)));

				switch (original_msg.type) {
				case COM_SAVE_FILE:
				case COM_SAVE_IMAGE_PPM:
				case COM_SAVE_IMAGE_TGA:
					if (data.data_return.error == EEXIST) {
#if FL_REVISION == 86
						char buf[1024];
						
						buf[sizeof buf - 1] = 0;
						sprintf(buf, "The file that you have specified:\n%s\nalready exists. Do you want to replace it ?", original_data.data_generic.data_string);
						if (buf[sizeof buf - 1] != 0) {
							abort_now();
						}
						
						if (fl_show_question(buf, -1)) {
#else /* 81 */
						if (fl_show_question("The file that you have specified:", original_data.data_generic.data_string, "already exists. Do you want to replace it ?")) {
#endif
							if (unlink(original_data.data_generic.data_string) == 0) {
								g2a_send(&Player, &original_msg, (const PMP_DATA_G2A *)original_data.data_generic.data_string);
							} else {
								show_alert("Cannot remove file:", original_data.data_generic.data_string, strerror(errno), 1);
							}
						}
					} else {
						show_alert("Cannot save in file:", original_data.data_generic.data_string, strerror(data.data_return.error), 1);
					}
					break;
				case COM_OPEN_FILE:
					show_alert("Cannot open file:", original_data.data_generic.data_string, strerror(data.data_return.error), 1);
					break;
				case COM_OPEN_FD:
					show_alert("Cannot play from Stdin:", "Stdin is not a pipe, file or socket", strerror(data.data_return.error), 1);
					break;
				default:
					show_alert("Error:", strerror(data.data_return.error), "", 1);
				}
			}
			break;
		default:
			/* msgs not understood should be ignored */
			printf("gui: received unknown msg (ignored): %s\n", pmp_print_msg(&msg, (const char *)&data));
			break;
	}

	return 0;
}


/*--------------------------------------------------------------------------+
| seek																		|
+--------------------------------------------------------------------------*/
static void
seek(n)
	int		n;
{
	if (Not_ready) {
		Pending_seek = n;
	} else {
		Pending_seek = -1;
		send_seek(n);
	}
}

/*--------------------------------------------------------------------------+
| warn_if_not_sync															|
+--------------------------------------------------------------------------*/
static void
warn_if_not_sync()
{
#if 0 /* Tristan Savatier - Fri Aug 8 17:31:27 PDT 1997 */
//	if (!fl_get_button(Sync->sync_sync)) {
//	
//		show_alert("Warning: audio is currently disabled.",
//				"Audio can be played only in Sync mode.",
//				"You can select Sync mode in Sync Options (Options->Sync...).",
//				1);
//				
//	} else
#endif /* 0 - Tristan Savatier - Fri Aug 8 17:31:27 PDT 1997 */
	if (!Framerate_is_nominal) {
	
		show_alert("Warning: audio is currently disabled.",
				"Audio can be played only when video plays at nominal framerate.",
				"Check Sync Options (Options->Sync...).",
				1);
	}
}

/*--------------------------------------------------------------------------+
| enter_fast_forward_mode													|
+--------------------------------------------------------------------------*/
static void
enter_fast_forward_mode()
{
	double frame_rate;
	
#if 0 /* Tristan Savatier - Sun Aug 10 00:45:55 PDT 1997 */
//	printf("Framerate_is_nominal = %d\n", Framerate_is_nominal);
#endif /* 0 - Tristan Savatier - Sun Aug 10 00:45:55 PDT 1997 */
	
	if (Framerate_is_nominal) {
		frame_rate = Nominal_framerate;
	} else {
		frame_rate = fl_get_slider_value(Sync->sync_framerate_slider);
	}
	
	Fast_forward = True;
	State_before_fast_forward = Gui_state;
	G2A_SEND_DOUBLE(COM_SYNC_FRAMERATE_SET, 0, FAST_FORWARD_COEF * frame_rate);
	
	if (State_before_fast_forward != STATE_PLAYING) {
		send_state(STATE_PLAYING);
	}
}

/*--------------------------------------------------------------------------+
| quit_fast_forward_mode													|
+--------------------------------------------------------------------------*/
static void
quit_fast_forward_mode()
{
	double frame_rate;

	if (!Fast_forward) {
		return;
	}

	if (Framerate_is_nominal) {
		frame_rate = Nominal_framerate;
	} else {
		frame_rate = fl_get_slider_value(Sync->sync_framerate_slider);
	}
	
	Fast_forward = False;
	G2A_SEND_DOUBLE(COM_SYNC_FRAMERATE_SET, 0, frame_rate);
	
	if (State_before_fast_forward != STATE_PLAYING) {
		send_state(STATE_PAUSED);
	}
}				

/*--------------------------------------------------------------------------+
| command																	|
+--------------------------------------------------------------------------*/
void
command(fl_object, param)
	FL_OBJECT *fl_object;
	long param;
{
	COMMAND	com = (COMMAND)param;
	
	switch (com) {
		case COM_QUIT:
			G2A_SEND_VOID(COM_QUIT, 0);
			exit(EXIT_SUCCESS);
			break;
		case COM_FILE_MENU:
			
			switch (fl_get_menu(Controls->file_menu)) {
				case FILE_MENU_OPEN:
					command(fl_object, COM_OPEN_FILE);
					break;
				case FILE_MENU_STDIN:
					command(fl_object, COM_OPEN_STDIN);
					break;
				case FILE_MENU_CLOSE:
					command(fl_object, COM_CLOSE_FILE);
					break;
				case FILE_MENU_SAVE_AS:
					command(fl_object, COM_SAVE_FILE);
					break;
				case FILE_MENU_INFOS:
					command(fl_object, COM_INFOS_OPEN);
					break;
				case FILE_MENU_QUIT:
					command(fl_object, COM_QUIT);
					break;
				case FILE_MENU_NONE:
				case FILE_MENU_HANGING:
					break;
				default:
					abort_now();
			}
			break;
		case COM_OPTIONS_MENU:
			switch (fl_get_menu(Controls->options_menu)) {
				case OPTIONS_MENU_AUDIO:
					command(fl_object, COM_AUDIO_OPEN);
					break;
				case OPTIONS_MENU_VIDEO:
					command(fl_object, COM_VIDEO_OPEN);
					break;
				case OPTIONS_MENU_SYNC:
					command(fl_object, COM_SYNC_OPEN);
					break;
				case OPTIONS_MENU_NONE:
				case OPTIONS_MENU_HANGING:
					break;
				default:
					abort_now();
			}
			break;
		case COM_WINDOW_MENU:
			switch (fl_get_menu(Controls->window_menu)) {
				case WINDOW_MENU_RAISE:
					command(fl_object, COM_WINDOW_RAISE);
					break;
				case WINDOW_MENU_FRAME_INFOS:
					command(fl_object, COM_FRAME_INFOS_OPEN);
					break;
				case WINDOW_MENU_SAVE_PPM:
					command(fl_object, COM_SAVE_IMAGE_PPM);
					break;
				case WINDOW_MENU_SAVE_TGA:
					command(fl_object, COM_SAVE_IMAGE_TGA);
					break;
				case WINDOW_MENU_NONE:
				case WINDOW_MENU_HANGING:
					break;
				default:
					abort_now();
			}
			break;
		case COM_HELP_MENU:
			switch (fl_get_menu(Controls->help_menu)) {
				case HELP_MENU_ABOUT:
					command(fl_object, COM_ABOUT_OPEN);
					break;
				case HELP_MENU_LICENSE:
					command(fl_object, COM_LICENSE_OPEN);
					break;
				case HELP_MENU_NONE:
				case HELP_MENU_HANGING:
					break;
				default:
					abort_now();
			}
			break;
		case COM_PLAY_PAUSE:
			reset_loop_if_pos_out_of_loop(Stream_pos);
			set_state(Gui_play_pause_set_state);
			break;
		case COM_STOP:
			reset_loop();
			set_state(STATE_STOPPED);
			break;
		case COM_STEP:
			update_state(STATE_PAUSED);			
			G2A_SEND_VOID(COM_STEP, 0);
			break;
		case COM_AUDIO_MUTE:
			{
				int 	muted = fl_get_button(fl_object);
				
				G2A_SEND_INT(com, 0, muted);
				if (!muted) {
					warn_if_not_sync();
				}
			}
			break;
		case COM_AUTO_EXIT:
		case COM_AUTO_PLAY:
		case COM_VIDEO_DITHER_COLORMAP_PRIVATE:
			G2A_SEND_INT(com, 0, fl_get_button(fl_object));
			break;
		case COM_LOOP:
			G2A_SEND_INT(com, 0, fl_get_button(fl_object));
			if (!fl_get_button(fl_object)) {
				reset_loop();
			}
			break;
		case COM_ZOOM:
			G2A_SEND_INT(com, 0, fl_get_button(fl_object) ? ZOOM_2X2 : ZOOM_1X1);
			break;
		case COM_AUDIO_CH_BOTH:
			G2A_SEND_INT(COM_AUDIO_CHANNEL, 0, AUDIO_CHANNEL_BOTH);
			break;
		case COM_AUDIO_CH_LEFT:
			G2A_SEND_INT(COM_AUDIO_CHANNEL, 0, AUDIO_CHANNEL_LEFT);
			break;
		case COM_AUDIO_CH_RIGHT:
			G2A_SEND_INT(COM_AUDIO_CHANNEL, 0, AUDIO_CHANNEL_RIGHT);
			break;
		case COM_AUDIO_CONTROLS_MASTER:
			G2A_SEND_INT(COM_AUDIO_CONTROLS, 0, AUDIO_CONTROLS_MASTER);
			break;
		case COM_AUDIO_CONTROLS_PCM:
#ifdef OPEN_SOUND_DRIVER
			G2A_SEND_INT(COM_AUDIO_CONTROLS, 0, AUDIO_CONTROLS_PCM);
#endif /* OPEN_SOUND_DRIVER */
			break;
		case COM_AUDIO_VOLUME:
			G2A_SEND_DOUBLE(com, 0, (double)fl_get_slider_value(fl_object));
			break;
		case COM_AUDIO_BALANCE:
			G2A_SEND_DOUBLE(com, 0, (double)fl_get_slider_value(fl_object));
			break;
		case COM_SEEK:
			seek((int)fl_get_slider_value(fl_object));
			break;
		case COM_REWIND:
			seek(0);
			get_position_parameters();
			break;
		case COM_END:
			reset_loop();
			G2A_SEND_VOID(COM_END, 0);
			get_position_parameters();
			break;
		case COM_FAST_FORWARD:
			{

				if (fl_get_button(fl_object)) {
					enter_fast_forward_mode();
				} else {
					quit_fast_forward_mode();
				}
			}
			break;
		case COM_ABOUT_OPEN:
			show_and_raise_form(About->about, FL_PLACE_SIZE, FL_FULLBORDER, "About MpegTV");
			break;
		case COM_ABOUT_CLOSE:
			fl_hide_form(About->about);
			break;
		case COM_LICENSE_OPEN:
			show_and_raise_form(License->license, FL_PLACE_CENTER | FL_FREE_SIZE, FL_FULLBORDER, "MpegTV Player 1.0 - License");
			break;
		case COM_LICENSE_CLOSE:
			fl_hide_form(License->license);
			break;
		case COM_AUDIO_OPEN:
			show_and_raise_form(Audio->audio, FL_PLACE_SIZE, FL_FULLBORDER, "MpegTV Audio");
			break;
		case COM_AUDIO_CLOSE:
			fl_hide_form(Audio->audio);
			break;
		case COM_VIDEO_OPEN:
			show_and_raise_form(Video->video, FL_PLACE_SIZE, FL_FULLBORDER, "MpegTV Video");
			break;
		case COM_VIDEO_CLOSE:
			fl_hide_form(Video->video);
			break;
		case COM_SYNC_OPEN:
			show_and_raise_form(Sync->sync, FL_PLACE_SIZE, FL_FULLBORDER, "MpegTV Sync");
			break;
		case COM_SYNC_CLOSE:
			fl_hide_form(Sync->sync);
			break;
		case COM_INFOS_OPEN:
			show_and_raise_form(Infos->infos, FL_PLACE_SIZE, FL_FULLBORDER, "MpegTV Sequence Infos");
			break;
		case COM_INFOS_CLOSE:
			fl_hide_form(Infos->infos);
			break;
		case COM_FRAME_INFOS_OPEN:
			show_and_raise_form(Frame_infos->frame_infos, FL_PLACE_SIZE, FL_FULLBORDER, "MpegTV Frame Infos");
			break;
		case COM_FRAME_INFOS_CLOSE:
			fl_hide_form(Frame_infos->frame_infos);
			break;
		case COM_ALERT_CLOSE:
			fl_hide_form(fl_object->form);
			fl_free_form(fl_object->form);
			break;
		case COM_OPEN_FILE:
			file_open();
			break;
		case COM_OPEN_STDIN:
			reset_auto_seek();
			reset_loop();
			G2A_SEND_INT(COM_OPEN_FD, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, 0);
			break;
		case COM_CLOSE_FILE:
			hide_stream_controls();
			G2A_SEND_VOID(COM_CLOSE_FILE, 0);
			break;
		case COM_SAVE_FILE:
			file_save();
			break;
		case COM_SAVE_IMAGE_PPM:
			image_save_ppm();
			break;
		case COM_SAVE_IMAGE_TGA:
			image_save_tga();
			break;
		case COM_VIDEO_DITHER_NONE:
			G2A_SEND_INT(COM_VIDEO_DITHER_TYPE, 0, DITHER_NONE);
			break;
		case COM_VIDEO_DITHER_GRAY:
			G2A_SEND_INT(COM_VIDEO_DITHER_TYPE, 0, DITHER_GRAY);
			break;
		case COM_VIDEO_DITHER_ORDERED:
			G2A_SEND_INT(COM_VIDEO_DITHER_TYPE, 0, DITHER_ORDERED);
			break;
		case COM_VIDEO_DITHER_COLOR:
			G2A_SEND_INT(COM_VIDEO_DITHER_TYPE, 0, DITHER_COLOR);
			break;
		case COM_VIDEO_GAMMA_SLIDER:
			if (Gamma.not_ready) {
				Gamma.pending = fl_get_slider_value(fl_object);
			} else {
				Gamma.not_ready = True;
				Gamma.pending = GAMMA_INVALID;
				G2A_SEND_DOUBLE(COM_VIDEO_GAMMA, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, (double)fl_get_slider_value(fl_object));
			}
			break;
		case COM_VIDEO_GAMMA_SLIDER_RESET:
			fl_set_slider_value(Video->video_gamma_slider, GAMMA_DEFAULT);
			command(Video->video_gamma_slider, COM_VIDEO_GAMMA_SLIDER);
			break;
		case COM_VIDEO_CONTRAST_SLIDER:
			if (Contrast.not_ready) {
				Contrast.pending = fl_get_slider_value(fl_object);
			} else {
				Contrast.not_ready = True;
				Contrast.pending = CONTRAST_INVALID;
				G2A_SEND_DOUBLE(COM_VIDEO_LUMA_FLOOR, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, - (double)fl_get_slider_value(fl_object));
			}
			break;
		case COM_VIDEO_CONTRAST_SLIDER_RESET:
			fl_set_slider_value(Video->video_contrast_slider, CONTRAST_DEFAULT);
			command(Video->video_contrast_slider, COM_VIDEO_CONTRAST_SLIDER);
			break;
		case COM_VIDEO_BRIGHTNESS_SLIDER:
			if (Brightness.not_ready) {
				Brightness.pending = fl_get_slider_value(fl_object);
			} else {
				Brightness.not_ready = True;
				Brightness.pending = BRIGHTNESS_INVALID;
				G2A_SEND_DOUBLE(COM_VIDEO_LUMA_SHIFT, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, (double)fl_get_slider_value(fl_object));
			}
			break;
		case COM_VIDEO_BRIGHTNESS_SLIDER_RESET:
			fl_set_slider_value(Video->video_brightness_slider, BRIGHTNESS_DEFAULT);
			command(Video->video_brightness_slider, COM_VIDEO_BRIGHTNESS_SLIDER);
			break;
		case COM_SYNC_I:
			/* fast forward disabled */
			if (Controls->play_pause->visible) {
				fl_hide_object(Controls->fast_forward);
			}
			G2A_SEND_INT(COM_SYNC_TYPE, 0, SYNC_I);
			Framerate_is_nominal = False;
			break;		
		case COM_SYNC_IP:
			/* fast forward disabled */
			if (Controls->play_pause->visible) {
				fl_hide_object(Controls->fast_forward);
			}
			G2A_SEND_INT(COM_SYNC_TYPE, 0, SYNC_IP);
			Framerate_is_nominal = False;
			break;		
		case COM_SYNC_IPB:
			/* fast forward disabled */
			if (Controls->play_pause->visible) {
				fl_hide_object(Controls->fast_forward);
			}
			G2A_SEND_INT(COM_SYNC_TYPE, 0, SYNC_IPB);
			Framerate_is_nominal = False;
			break;		
		case COM_SYNC_NOMINAL:
			/* fast forward enabled */
			if (Controls->play_pause->visible) {
				fl_show_object(Controls->fast_forward);
			}
			G2A_SEND_DOUBLE(COM_SYNC_FRAMERATE_SET, 0, Nominal_framerate);
			Framerate_is_nominal = True;
			G2A_SEND_INT(COM_SYNC_TYPE, 0, SYNC_FRAMERATE);
			break;		
		case COM_SYNC_SYNC:
#if 0 /* Tristan Savatier - Thu Aug 28 20:51:56 PDT 1997 */
//			/* fast forward enabled */
//			if (Controls->play_pause->visible) {
//				fl_show_object(Controls->fast_forward);
//			}
//			G2A_SEND_INT(COM_SYNC_TYPE, 0, SYNC_FRAMERATE);
//			if (Specified_framerate_is_nominal) {
//				G2A_SEND_DOUBLE(COM_SYNC_FRAMERATE_SET, 0, Nominal_framerate);
//				Framerate_is_nominal = True;
//			} else {
//				G2A_SEND_DOUBLE(COM_SYNC_FRAMERATE_SET, 0, (double)fl_get_slider_value(Sync->sync_framerate_slider));
//				Framerate_is_nominal = False;
//			}
#endif /* 0 - Tristan Savatier - Thu Aug 28 20:51:56 PDT 1997 */
			set_sync_sync();
			break;		
		case COM_SYNC_FRAMERATE_SET:
#if 0 /* Tristan Savatier - Thu Aug 28 20:58:33 PDT 1997 */
//			set_framerate();
#endif /* 0 - Tristan Savatier - Thu Aug 28 20:58:33 PDT 1997 */
			Specified_framerate_is_nominal = False;
			if (fl_get_button(Sync->sync_sync)) {
				G2A_SEND_DOUBLE(com, 0, (double)fl_get_slider_value(fl_object));
				Framerate_is_nominal = False;
			}
			break;
		case COM_SYNC_FRAMERATE_RESET:
			fl_set_slider_value(Sync->sync_framerate_slider, Nominal_framerate);
			Specified_framerate_is_nominal = True;
			if (fl_get_button(Sync->sync_sync)) {
				G2A_SEND_DOUBLE(COM_SYNC_FRAMERATE_SET, 0, Nominal_framerate);
				Framerate_is_nominal = True;
				G2A_SEND_INT(COM_SYNC_TYPE, 0, SYNC_FRAMERATE);
			}
			break;
		case COM_AUDIO_QUALITY_LOW:
			G2A_SEND_INT(COM_AUDIO_QUALITY, 0, AUDIO_QUALITY_LOW);
			break;
		case COM_AUDIO_QUALITY_MEDIUM:
			G2A_SEND_INT(COM_AUDIO_QUALITY, 0, AUDIO_QUALITY_MEDIUM);
			break;
		case COM_AUDIO_QUALITY_HIGH:
			G2A_SEND_INT(COM_AUDIO_QUALITY, 0, AUDIO_QUALITY_HIGH);
			break;
		case COM_AUDIO_PORT_LINE_OUT:	/* FALLTHROUGH */
		case COM_AUDIO_PORT_HEADPHONE:	/* FALLTHROUGH */
		case COM_AUDIO_PORT_SPEAKER:
#if 0 /* Tristan Savatier - Sat Aug 16 18:50:34 PDT 1997 */
//			if (Audio_ports == 0) {
//				fl_set_button(Audio->audio_port_line_out, 1);
//			} else
#endif /* 0 - Tristan Savatier - Sat Aug 16 18:50:34 PDT 1997 */
			{
				int		ports = 0;
				if (fl_get_button(Audio->audio_port_line_out)) {
					ports |= AUDIO_PORT_LINE_OUT;
				}
				if (fl_get_button(Audio->audio_port_headphone)) {
					ports |= AUDIO_PORT_HEADPHONE;
				}
				if (fl_get_button(Audio->audio_port_speaker)) {
					ports |= AUDIO_PORT_SPEAKER;
				}
				G2A_SEND_INT(COM_AUDIO_PORTS_SELECT, 0, ports);
			}			
			break;
		default:
			G2A_SEND_VOID(com, 0);
			break;
	}

	/*
	 * After certain commands, we prefer to update the position parameters immediately,
	 * rather than waiting for the automatic update that takes place every
	 * SLIDER_UPDATE_FREQUENCY seconds.
	 */
	switch (com) {
		case COM_CLOSE_FILE:	/* FALLTHROUGH */
		case COM_STATE:			/* FALLTHROUGH */
		case COM_STEP:			/* FALLTHROUGH */
			get_position_parameters();
			break;
		default:
			break;
	}
}

/*--------------------------------------------------------------------------+
| user_options_set_default													|
+--------------------------------------------------------------------------*/
static void
user_options_set_default()
{	
	User_options.loop = False;
	User_options.zoom = ZOOM_1X1;
	User_options.auto_exit = False;
	User_options.auto_play = True;
	User_options.dither_type = (DITHER_TYPE)-1;
	User_options.sync_type = SYNC_FRAMERATE;
	User_options.msg_select = 0;
	User_options.audio_msg_select = 0;
	User_options.port = 0;
	User_options.port_udp = 0;
	User_options.stdin_mode = False;
	User_options.gamma = GAMMA_DEFAULT;
	User_options.contrast = CONTRAST_DEFAULT;
	User_options.brightness = BRIGHTNESS_DEFAULT;
	User_options.audio_mode = AUDIO_MODE_PLAY;
	User_options.audio_quality = AUDIO_QUALITY_MEDIUM;
	User_options.audio_channel = AUDIO_CHANNEL_LEFT;
	User_options.audio_controls = AUDIO_CONTROLS_PCM;
	User_options.audio_pcm_volume = INITIAL_PCM_VOLUME;
	User_options.audio_pcm_balance = INITIAL_PCM_BALANCE;
	User_options.audio_master_volume = INITIAL_MASTER_VOLUME;
	User_options.audio_master_balance = INITIAL_MASTER_BALANCE;
	User_options.audio_ports = -1;
	User_options.xforms_argv = NULL;
	User_options.pmp_base_fd = 3;
	User_options.existing_window = -1;
	User_options.loop_start_pos = 0;
	User_options.loop_end_pos = 0;
	User_options.mtvp_geometry_x = 200;
	User_options.mtvp_geometry_y = 300;
	User_options.geometry_x = 200;
	User_options.geometry_y = 192+80-95;
	User_options.geometry_w = 490;
	User_options.geometry_h = 95;
}


/*--------------------------------------------------------------------------+
| oneliner_show																|
+--------------------------------------------------------------------------*/
static void
oneliner_show()
{
	FL_OBJECT *ob = Oneliner_object;
	
	if (ob != NULL && ob->u_vdata != NULL && ob->visible) {
		fl_show_oneliner((const char *)ob->u_vdata, ob->form->x + ob->x + ob->w / 2,
                        ob->form->y + ob->y + ob->h + 5);
	}
}

/*--------------------------------------------------------------------------+
| post_handler																|
+--------------------------------------------------------------------------*/
static int
post_handler(FL_OBJECT *ob, int event,
              FL_Coord mx, FL_Coord my, int key, void *xev)
{
	if(event == FL_ENTER) {
     	Oneliner_object = ob;
		fl_set_timer(Controls->oneliner_timer, 1.0);
	} else if (Oneliner_object == ob && (event == FL_PUSH
										|| event == FL_RELEASE
										|| event == FL_LEAVE
										|| event == FL_MOUSE
										|| event == FL_UNFOCUS
										|| event == FL_KEYBOARD
										|| event == FL_STEP
										|| event == FL_SHORTCUT
										|| event == FL_FREEMEM
										|| event == FL_OTHER
										|| event == FL_DBLCLICK
										|| event == FL_TRPLCLICK
										|| event == FL_PS)) {
		fl_set_timer(Controls->oneliner_timer, 0.0);
		Oneliner_object = NULL;
		fl_hide_oneliner();
	}
	
	/* Work around a bug of the IN-OUT buttons in Xforms! */
	if (Fast_forward && ob == Controls->fast_forward && event == FL_LEAVE) {
		quit_fast_forward_mode();
	}
	
	if (ob == Controls->frame_slider) {		
		if (event == FL_PUSH) {
			Seeking = True;
			State_before_seeking = Gui_state;
			send_state(STATE_PAUSED);
			if (State_before_seeking == STATE_STOPPED) {
				update_state(STATE_PAUSED);
			}
			
			fl_set_object_label(Controls->actual_framerate, "n/a");
			fl_set_object_label(Controls->number_of_frames_played, "0");
		} else if (event == FL_RELEASE) {
			Seeking = False;
			if (State_before_seeking == STATE_PLAYING) {
				send_state(STATE_PLAYING);
			}
		}
	}

	return 0;
}

/*--------------------------------------------------------------------------+
| set_oneliners																|
+--------------------------------------------------------------------------*/
static void
set_oneliners()
{
	fl_mapcolor(FL_FREE_COL1, 255, 250, 205);
	fl_set_oneliner_color(FL_BLACK, FL_FREE_COL1);
	
	SET_ONELINER(Controls->play_pause, "Play/Pause");
	SET_ONELINER(Controls->step, "Step to next frame");
	SET_ONELINER(Controls->stop, "Stop");
	SET_ONELINER(Controls->time_total, "Total play time HH:MM:SS");
	SET_ONELINER(Controls->time_elapsed, "Elapsed play time HH:MM:SS");
	SET_ONELINER(Controls->end, "Go to end");
	SET_ONELINER(Controls->actual_framerate, "Actual framerate\nin frames per second");
	SET_ONELINER(Controls->number_of_frames_played, "Number of frames played\nsince last random access");
	SET_ONELINER(Controls->loop, "Auto-repeat on/off");
	SET_ONELINER(Controls->audio_mute, "Audio on/off");
	SET_ONELINER(Controls->zoom, "Zoom (2x2) on/off");
	SET_ONELINER(Controls->fast_forward, "Fast forward");
	SET_ONELINER(Controls->rewind, "Rewind");
	SET_ONELINER(Controls->auto_play, "Auto-play on/off. When on, starts playing\n automatically when opening a file");
	SET_ONELINER(Controls->frame_slider, "Move to any position in the sequence");

#ifdef OPEN_SOUND_DRIVER
	SET_ONELINER(Controls->audio_volume_slider, "If audio mixer is available,\nVolume and Balance controls\ncan be attached to PCM or Master levels\n(see Audio options)");
	SET_ONELINER(Controls->audio_balance_slider, "If audio mixer is available,\nVolume and Balance controls\ncan be attached to PCM or Master levels\n(see Audio options)");
#endif
	
	SET_ONELINER(Audio->audio_controls_pcm, "Attach Volume and Balance controls\nto PCM levels");
	SET_ONELINER(Audio->audio_controls_master, "Attach Volume and Balance controls\nto Master levels (if audio mixer is available)");
	SET_ONELINER(Audio->audio_quality_low, "ulaw, if supported (8 KHz)");
	SET_ONELINER(Audio->audio_quality_medium, "Half sampling freq. (22 KHz)");
	SET_ONELINER(Audio->audio_quality_high, "Nominal sampling freq. (44 KHz)");
	SET_ONELINER(Audio->audio_port_line_out, "Line-out jack");
	SET_ONELINER(Audio->audio_port_headphone, "Headphone jack");
	SET_ONELINER(Audio->audio_port_speaker, "Built-in speaker");
	SET_ONELINER(Audio->audio_ch_both, "Both channels are played only\nwhen audio is coded in stereo\nor in dual-channel mode");
	SET_ONELINER(Audio->audio_ch_right, "Right channel is played only\nwhen audio is coded in stereo\nor in dual-channel mode");

	SET_ONELINER(Infos->mux_bitrate, "Total bitrate of multiplexed stream\n(Video + Audio + Systems)");
	SET_ONELINER(Infos->video_framerate, "Nominal framerate");
	SET_ONELINER(Infos->audio_frequency, "Nominal sampling frequency");
	SET_ONELINER(Infos->video_constrained, "Indicates if the constrained-parameter\nflag is set");
	SET_ONELINER(Infos->video_compliant, "Indicates if this is a compliant\nconstrained-parameter MPEG-1 video bitstream");

	SET_ONELINER(Sync->sync_i, "Play only I-frames\nas fast as possible (no audio)");	
	SET_ONELINER(Sync->sync_ip, "Play only I- and P-frames\nas fast as possible (no audio)");	
	SET_ONELINER(Sync->sync_ipb, "Play all video frames\nas fast as possible (no audio)");	
	SET_ONELINER(Sync->sync_sync, "Play video at the specified framerate\nNote: Audio can be played only at Nominal framerate");
	SET_ONELINER(Sync->sync_framerate_slider, "Select video framerate");	
	SET_ONELINER(Sync->sync_reset_framerate_slider, "Reset video framerate to Nominal framerate\nAudio can be played only at Nominal framerate");
	
}


/*--------------------------------------------------------------------------+
| do_gui_fork																|
+--------------------------------------------------------------------------*/
static void
do_gui_fork(argv0, argv)
	const char *argv0;
	const char **argv;
{
	int		argc;
	FL_OBJECT *fl_object;
	static const char	*fl_argv[] = {
		/* app name updated later */
		NULL,
		
		/* border width -1 looks better! */
		"-bw",
		"-1",
		
		/* try to work-around the famous Xforms clipped-menus bug: */
#ifdef SGI		
		"-visual",
		"Default",
		"-depth",
		"8",
#else
		"-visual",
		"PseudoColor",
#endif
		/* probably a good idea */
		"-shared",

		NULL,
	};
	
	if (argv == NULL) {
		argv = fl_argv;
	}
	argv[0] = argv0;
	{
		const char **p = argv;
		
		for (argc = 0; p[argc] != NULL; argc++);
	}
	
	if (fl_initialize (&argc, (char **)argv, "FormDemo", 0, 0) == NULL) {
		FATAL_ERROR(("cannot initialize xform"));
	}
	

	{
#include "mpegtv.xbm"
		Icon_pixmap = fl_create_from_bitmapdata(fl_default_win(), mpegtv_bits, mpegtv_width, mpegtv_height);
	}

	fl_set_idle_callback(my_idle, NULL);
	
	if ((Controls = create_form_controls()) == NULL) {
		FATAL_ERROR(("cannot create Controls form"));
	}
	
	fl_set_form_atclose(Controls->controls, at_close_controls, NULL);
	fl_set_form_icon(Controls->controls, Icon_pixmap, 0);
	
	if ((About = create_form_about()) == NULL) {
		FATAL_ERROR(("cannot create About form"));
	}
	fl_set_form_atclose(About->about, at_close_hide, NULL);
	fl_set_form_icon(About->about, Icon_pixmap, 0);

	if ((License = create_form_license()) == NULL) {
		FATAL_ERROR(("cannot create License form"));
	}
	fl_set_form_atclose(License->license, at_close_hide, NULL);
	fl_set_form_icon(License->license, Icon_pixmap, 0);

	if ((Audio = create_form_audio()) == NULL) {
		FATAL_ERROR(("cannot create Audio form"));
	}
	fl_set_form_atclose(Audio->audio, at_close_hide, NULL);
	fl_set_form_icon(Audio->audio, Icon_pixmap, 0);
	
	if ((Video = create_form_video()) == NULL) {
		FATAL_ERROR(("cannot create Video form"));
	}
	fl_set_form_atclose(Video->video, at_close_hide, NULL);
	fl_set_form_icon(Video->video, Icon_pixmap, 0);
	
	if ((Sync = create_form_sync()) == NULL) {
		FATAL_ERROR(("cannot create Sync form"));
	}
	fl_set_form_atclose(Sync->sync, at_close_hide, NULL);
	fl_set_form_icon(Sync->sync, Icon_pixmap, 0);
	
	if ((Infos = create_form_infos()) == NULL) {
		FATAL_ERROR(("cannot create Infos form"));
	}
	fl_set_form_atclose(Infos->infos, at_close_hide, NULL);
	fl_set_form_icon(Infos->infos, Icon_pixmap, 0);
	
	if ((Frame_infos = create_form_frame_infos()) == NULL) {
		FATAL_ERROR(("cannot create Frame_infos form"));
	}
	fl_set_form_atclose(Frame_infos->frame_infos, at_close_hide, NULL);
	fl_set_form_icon(Frame_infos->frame_infos, Icon_pixmap, 0);
	
	{
		u_int		i;
		for (i = 0; i < sizeof License_lines / sizeof *License_lines; i++) {
			fl_add_browser_line(License->browser, License_lines[i]);
		}
		fl_set_browser_fontstyle(License->browser, FL_FIXED_STYLE);
	}
	
	fl_set_menu(Controls->file_menu, "Open File...|Play from Stdin|Close|Save as...|Sequence infos%l|Quit");

	fl_set_menu(Controls->options_menu, "Audio...|Video...|Framerate & Sync...");

	fl_set_menu(Controls->window_menu, "Raise window|Frame infos|Capture Frame (PPM format)...|Capture Frame (TGA format)...");

	fl_set_menu(Controls->help_menu, "About...|License...");

	fl_set_slider_bounds(Sync->sync_framerate_slider, FRAMERATE_MIN, FRAMERATE_MAX);
	fl_set_slider_bounds(Controls->audio_volume_slider, 0.0, 1.0);

	fl_set_slider_bounds(Controls->audio_balance_slider, 0.0, 1.0);
	fl_set_slider_size(Controls->frame_slider, 0.04);

	fl_set_slider_bounds(Video->video_gamma_slider, GAMMA_MIN, GAMMA_MAX);
	fl_set_slider_size(Video->video_gamma_slider, 0.04);
	fl_set_slider_value(Video->video_gamma_slider, User_options.gamma);
	Gamma.pending = GAMMA_INVALID;
	command(Video->video_gamma_slider, COM_VIDEO_GAMMA_SLIDER);
	
	fl_set_slider_bounds(Video->video_contrast_slider, CONTRAST_MIN, CONTRAST_MAX);
	fl_set_slider_size(Video->video_contrast_slider, 0.04);
	fl_set_slider_value(Video->video_contrast_slider, User_options.contrast);
	Contrast.pending = CONTRAST_INVALID;
	command(Video->video_contrast_slider, COM_VIDEO_CONTRAST_SLIDER);

	fl_set_slider_bounds(Video->video_brightness_slider, BRIGHTNESS_MIN, BRIGHTNESS_MAX);
	fl_set_slider_size(Video->video_brightness_slider, 0.04);
	fl_set_slider_value(Video->video_brightness_slider, User_options.brightness);
	Brightness.pending = BRIGHTNESS_INVALID;
	command(Video->video_brightness_slider, COM_VIDEO_BRIGHTNESS_SLIDER);
	
	/*fl_set_form_geometry(Controls->controls, 200, 192+80-95, 490, 95);*/
	fl_set_form_geometry(Controls->controls,
							(FL_Coord)User_options.geometry_x,
							(FL_Coord)User_options.geometry_y,
							(FL_Coord)User_options.geometry_w,
							(FL_Coord)User_options.geometry_h);

	hide_stream_controls();
	fl_hide_object(Video->video_dither_none);
	fl_hide_object(Video->video_dither_gray);
	fl_hide_object(Video->video_dither_ordered);
	fl_hide_object(Video->video_dither_ordered_colormap_private);
	fl_hide_object(Video->video_dither_color);

	fl_hide_object(Audio->audio_port_line_out);
	fl_hide_object(Audio->audio_port_headphone);
	fl_hide_object(Audio->audio_port_speaker);

#if 0 /* Tristan Savatier - Mon Aug 11 02:50:51 PDT 1997 */
//	if (User_options.existing_window == -1) {
#endif /* 0 - Tristan Savatier - Mon Aug 11 02:50:51 PDT 1997 */
		show_and_raise_form(Controls->controls, User_options.iconify ? FL_PLACE_ICONIC : FL_PLACE_POSITION, FL_FULLBORDER, "MpegTV Controls");
#if 0 /* Tristan Savatier - Mon Aug 11 02:50:56 PDT 1997 */
//	}
#endif /* 0 - Tristan Savatier - Mon Aug 11 02:50:56 PDT 1997 */

	fl_add_io_callback(Player.receive_channel.msg_pipe[PMP_FD_READ], FL_READ, (void (*)(int, void *))a2g_read_command, NULL);

	fl_set_object_label(About->version, VERSION_STR);
	
	set_oneliners();

	gui_set_user_options();

	while(True) {
#ifdef MSG_DEBUG
		printf("calling fl_do_forms\n");
#endif

		if ((fl_object = fl_do_forms()) == Controls->oneliner_timer) {
			/* fl_do_forms() will return the oneliner_timer when it expires */
			oneliner_show();
		} else {
			/* showl not happen */
			printf("Error: fl_do_forms() returned!\n");
		}
	}
}

#ifndef BSDI

/* for some reason this does not work with BSDI! */

/*--------------------------------------------------------------------------+
| my_own_error_handler														|
+--------------------------------------------------------------------------*/
static void
my_own_error_handler(const char *func, const char *fmt,...)
{
	/* our dummy Xforms warning message routine does nothing */
}

/*--------------------------------------------------------------------------+
| whereError																|
+--------------------------------------------------------------------------*/
void
(*whereError(int something,
                  int level,
                  const char *string,
                  int some_int))(const char *func, const char *fmt,...)
{
	/* whereError() shadows Xforms warning message routine */

    return my_own_error_handler;
}

#endif /* not BSDI */

/*--------------------------------------------------------------------------+
| handler_ignore															|
+--------------------------------------------------------------------------*/
/*ARGSUSED*/
static void
handler_ignore(sig)
	int     sig;
{
	/*
	 * Caution: never install a handler like handler_ignore
	 * for catching SIGCHLD on a HP-UX system. A handler that
	 * catches SIGCHLD on HP-UX should always call wait(),
	 * wait3() or waitpid() before reinstalling itself!
	 * See "man 5 signal" and look for SIGCLD.
	 */
	WARNING(("(pid=%u): received signal %d (ignored)", (u_int)Pid, sig));

#ifdef SIGNAL_RESETS_HANDLER
	(void)signal(sig, handler_ignore);
#endif
}


/*--------------------------------------------------------------------------+
| clean_exit																|
+--------------------------------------------------------------------------*/
void
clean_exit(n)
	int		n;
{
	WARNING(("(pid=%u): exiting, status = %d", (u_int)Pid, n));

	/* make sure we ignore SIGCHLD signals */
	(void)signal(SIGCHLD, SIG_IGN);
	
	/* send SIGTERM signal to Player */
	if (Player.pid != 0) {
		kill(Player.pid, SIGTERM);
	}
	
	/* terminate (real exit this time) */
	exit(n);
}

/*--------------------------------------------------------------------------+
| handler_exit																|
+--------------------------------------------------------------------------*/
/*ARGSUSED*/
static void
handler_exit(sig)
	int     sig;
{
	WARNING(("(pid=%u): received signal %d (aborting)", (u_int)Pid, sig));

	clean_exit(EXIT_FAILURE);
}


/*--------------------------------------------------------------------------+
| handler_sigchld															|
+--------------------------------------------------------------------------*/
/*ARGSUSED*/
static void
handler_sigchld(sig)
	int     sig;
{
	/*
	 * Note: If you want to call any Xlib or Xforms routine from
	 * a signal handler, you must register it using
	 * fl_add_signal_callback() so that
	 * the signal handler will be executed in the proper context
	 * in the Xforms event loop.  Please refer to the Xforms
	 * manual, Signals section.
	 *
	 * This particular signal handler does not
	 * call any Xlib or Xforms routine, so it is safe
	 * to install it with signal().
	 */

	WARNING(("(pid=%u): received signal %d", (u_int)Pid, sig));

	if (Player.pid != 0) {
	
		/* check if player actually died */
		switch (g2a_child_died(&Player)) {
		case 0:
			WARNING(("(pid=%u): child process %u still alive, signal ignored", (u_int)Pid, (u_int)Player.pid));
			break;
		case 1:
			WARNING(("(pid=%u): child process %u still alive, signal ignored", (u_int)Pid, (u_int)Player.pid));
			EXIT(EXIT_FAILURE);
		case -1:
			FATAL_ERROR(("(pid=%u): g2a_child_died(%u) error", (u_int)Pid, (u_int)Player.pid));
		default:
			FATAL(("What?!? g2a_child_died() returned bad value!"));
		}			
	}

#ifdef SIGNAL_RESETS_HANDLER
	(void)signal(sig, handler_sigchld);
#endif
}


/*--------------------------------------------------------------------------+
| init_signals																|
+--------------------------------------------------------------------------*/
static void
init_signals()
{
	portable_catch_termination_signals(handler_exit);

	(void)signal(SIGCHLD, handler_sigchld);
}


/*--------------------------------------------------------------------------+
| get_int0																	|
+--------------------------------------------------------------------------*/
static int
get_int0(o, p, vmin, vmax)
	char	o;
	const char	**p;	/* side effect on purpose: *p will be updated */
	int	vmin, vmax;
{
	const char	*s = *p;
	int		n;

	n = strtol(s, (char **)p, 0);
	if (*p == s) {
		FATAL(("bad option %c%s (number expected)", o, s));
	}
	if (vmin <= vmax && (n < vmin || n > vmax)) {
		FATAL(("bad option %c%d (%d out of range [%d, %d])", o, n, n, vmin, vmax));
	}
	return n;
}


/*--------------------------------------------------------------------------+
| get_int																	|
+--------------------------------------------------------------------------*/
static int
get_int(o, s, vmin, vmax)
	char	o;
	const char	*s;
	int		vmin, vmax;
{
	return get_int0(o, &s, vmin, vmax);
}

/*--------------------------------------------------------------------------+
| get_double																|
+--------------------------------------------------------------------------*/
static double
get_double(o, s, vmin, vmax)
	char	o;
	const char	*s;
	double		vmin, vmax;
{
	char	*p;
	double	d;

	d = strtod(s, &p);
	if (p == s) {
		FATAL(("bad option %c%s (floating number expected)", o, s));
	}

	if (vmin <= vmax && (d < vmin || d > vmax)) {
		FATAL(("bad option %c%4.2f (%4.2f out of range [%4.2f, %4.2f])", o, d, d, vmin, vmax));
	}

	return d;
}


/*--------------------------------------------------------------------------+
| usage - print the help.													|
+--------------------------------------------------------------------------*/
static void
usage()
{
	static const char *helptab[] = {
"MpegTV Player 1.0 - %s\n",
"Copyright (c) 1997, MpegTV LLC\n",
"ALL RIGHTS RESERVED\n",
"\n",
"Portions of this software:\n",
"Copyright (c) 1995, 1996 by T.C. Zhao and Mark Overmars\n",
"ALL RIGHTS RESERVED\n",
"\n",
"Please read the LICENSE file. Use at your own risks!\n",
"MpegTV Home page: http://www.mpegtv.com\n",
"\n",
"Function:  Graphic front-end for the MpegTV MPEG Player.\n",
"           The MpegTV MPEG Player (mtvp) is a realtime\n",
"           software-only MPEG Video player with audio/sync\n",
"Note:      mtv requires mtvp, the MpegTV software MPEG Player\n",
"Usage:     mtv {options} [file]\n",
"           file may be a file or '-' for stdin\n",
"           file may also be directory\n",
"Current limitations:\n",
"           Decode only MPEG-1 Systems or Video streams\n",
"           Play only elementary audio and video streams #0\n",
"           GUI does not allow to control all the parameters\n",
"Options:   Single letter options with no parameters can be combined\n",
"           Example: -lz.\n",
"           One option with parameters can be appended to a list of\n",
"           options with no parameters. Example: -wxd2\n",
" -P<n>     read input from socket (SOCK_STREAM) on port <n>\n",
" -d<n>     display mode:\n",
"           n=0: no display (but frames are decoded and audio is played)\n",
"           n=1: gray\n",
"           n=2: 8-bit pseudo-color\n",
"           n=3: 16-bit true color\n",
"           By default mtv picks the best display mode for your display.\n",
" -s<n>     synchronization type:\n",
"           Note that audio is played only with -s0 (default)\n",
"           n=0: synchronize to match sequence nominal frame-rate (default)\n",
"           n=1: display all frames as fast as possible\n",
"           n=2: display all I and P frames as fast as possible\n",
"           n=3: display all I frames as fast as possible\n",
" -sf<n>    play video at specified framerate [1.0 to 60.0]\n",
" -g<n>     gamma correction, range [0.1 to 2.0]\n",
"           (default: irix: 0.8, solaris: 1.4, other: 1.0)\n",
"           n large produces lighter images, n small produces darker images\n",
" -c<n>     contrast, range [-0.50 to 0.50] (default 0.0)\n",
"           n > 0 for more contrast, n < 0 for less contrast\n",
" -b<n>     brightness, range [-0.50 to 0.50] (default 0.0)\n",
"           n > 0 for brighter, n < 0 for darker\n",
" -w[<n>]   print informative messages and warnings.\n",
"           Several values can be specified with several -w<n> options.\n",
"           If <n> is not specified all the messages are printed.\n",
"           n=0: general messages\n",
"           n=1: non-fatal error messages\n",
"           n=2: warnings (e.g. when errors are found in MPEG stream)\n",
"           n=3: frame sync messages\n",
"           n=4: audio PTS messages\n",
"           n=5: interprocess communication messages\n",
"           n=6: input stream read messages\n",
"           n=7: systems parsing messages\n",
"           n=8: Xlib messages\n",
"           n<0: set message mask to -n - 1\n",
" -l        loop (only works when reading from a file)\n",
" -z        zoom (2x2)\n",
" -i        iconify controls\n",
" -G[=]{+-}<x>{+-}<y>  specify initial position of player window\n",
" -F[=][<w>{xX}<h>]{+-}<x>{+-}<y>  specify initial position and size\n",
"           of control window\n",
" -a<n>     audio mode\n",
"           n=0: no audio\n",
"           n=1: play audio (default)\n",
"           n=2: output audio stream on stdout\n",
" -aw[<n>]  audio process informative messages and warnings.\n",
"           Several values can be specified with several -aw<n> options.\n",
"           If <n> is not specified all the messages are printed.\n",
"           n=0: general messages\n",
"           n=1: non-fatal error messages\n",
"           n=2: warnings (e.g. when errors are found in MPEG stream)\n",
"           n=4: audio PTS messages\n",
"           n=5: interprocess communication messages\n",
"           n=8: audio sync messages\n",
"           n=9: audio ioctl messages\n",
"           n=10: audio write sample messages\n",
"           n<0: set message mask to -n - 1\n",
" -aq<n>    audio quality\n",
"           n=0: low\n",
"           n=1: medium (default)\n",
"           n=2: high\n",
" -ac<n>    audio channel\n",
"           n=0: both (stereo)\n",
"           n=1: left (default)\n",
"           n=2: right\n",
#ifdef SOLARIS
" -ao<n>    audio output port(s)\n",
"           This option can be used several times to select more than one port.\n",
"           By default player uses ports currently selected.\n",
"           n=0: internal speaker\n",
"           n=1: headphone jack\n",
"           n=2: line-out\n",
#endif
#ifdef OPEN_SOUND_DRIVER
" -aM       volume and balance controls are linked to MASTER mixer volume\n",
"           by default volume and balance controls are linked to PCM volume\n",
#endif
" -a<c><x><n> set initial audio parameter\n",
"           n is a double in the range [0.0, 1.0] or -1 to keep the setting unchanged\n",
"           c=p: PCM\n",
"           c=m: master (only on systems with a /dev/mixer device)\n",
"           x=v: volume\n",
"           x=b: balance (n=0.0 means left, n=0.5 means centered, n=1.0 means right)\n",
"           By default, options -amv1.0 -amb0.5 is assumed, i.e. master volume is set to\n",
"           100% at startup. You can use -amv-1 -amb-1 to keep the current setting.\n",
" -p        auto-play off (default auto-play on)\n",
"           Note: auto-play is always on when non-interactive\n",
" -e        auto-exit on (default auto-exit off)\n",
" -W<n>     specify window id.  E.g. you can give the window id of the root window.\n",
"           n=0: create new window (default)\n",
"           n>0: play in an existing window.\n",
" -S<n>     auto-seek byte offset position (default 0). Player will seek to this\n",
"           position right after opening the specified file. Can be used with -p.\n",
" -E<n>     auto-end byte offset position (default end-of-file).\n",
"           When the auto-end position is reached:\n",
"           If Loop on, Player will seek back to auto-seek position and loop.\n",
"           If Loop off, Player will pause after passing this byte offset,\n",
"           or it will exit if auto-exit is on.\n",
"           Auto-seek and auto-end are only valid for the file specified in the\n",
"           command-line and are reset to defaults when user seeks out of the\n",
"           interval [auto-seek, auto-end] or turns Loop off.\n",
" -x<name>  name of MpegTV MPEG Player executable file (default mtvp)\n",
" +         all the options and parameters following a single + option\n",
"           are passed to the Xforms library.  See Xforms manual for infos.\n",

	NULL,
	};
	const char	**p = helptab;

	while (*p != NULL) {
		printf(*p++, VERSION_STR);
	}
	printf("Please send bug reports and comments to tristan@mpegtv.com\n");
}

/*--------------------------------------------------------------------------+
| parse_options - parse the option string "p".								|
| The first "-" character is assumed to be skipped before this function		|
| is called.																|
| This function is called when an unknown option is encountered.			|
+--------------------------------------------------------------------------*/
static void
parse_options(p)
	const char *p;
{
	int		o;

	for (; ; ) {
		switch (o = *p++) {
		case 'h':	usage(); EXIT(EXIT_FAILURE); continue;
		case 'w':	if (!isdigit(*p)) {
						User_options.msg_select = (u_int)0x7fffffff;
						Msg_select = (u_int)0x7fffffff;
						continue;
					} else {
						User_options.msg_select |= (1 << get_int(o, p, 0, 31));
						Msg_select |= (1 << get_int(o, p, 0, 31));
					}
					return;
		case 'l':	User_options.loop = True; continue;
		case 'z':	User_options.zoom = ZOOM_2X2; continue;
		case 'i':	User_options.iconify = True; continue;
		case 'G':	{
					u_int	w, h;
					XParseGeometry(p,
						&User_options.mtvp_geometry_x,
						&User_options.mtvp_geometry_y,
						&w,
						&h);
				}
				return;
		case 'F':	XParseGeometry(p,
						&User_options.geometry_x,
						&User_options.geometry_y,
						&User_options.geometry_w,
						&User_options.geometry_h);
				return;
		case 'a':
			if (isdigit(*p)) {
				User_options.audio_mode = (AUDIO_MODE)get_int(o, p, 0, (int)AUDIO_MODE_COUNT - 1);
				return;
			} else {
				switch (o = *p++) {
				case 'w':
					if (!isdigit(*p) && !(*p == '-' && isdigit(*(p + 1)))) {
						User_options.audio_msg_select = (u_int)0x7fffffff;
						continue;
					} else {
						int		n = get_int(o, p, INT_MIN, 31);
						
						if (n >= 0) {
							User_options.audio_msg_select |= (1 << n);
						} else {
							User_options.audio_msg_select = - (n + 1);
						}
					}
					break;
				case 'q':
					User_options.audio_quality = (AUDIO_QUALITY)get_int(o, p, 0, (int)AUDIO_QUALITY_COUNT - 1);
					break;
				case 'c':
					User_options.audio_channel = (AUDIO_CHANNEL)get_int(o, p, 0, (int)AUDIO_CHANNEL_COUNT - 1);
					break;
#ifdef OPEN_SOUND_DRIVER
				case 'M':
					User_options.audio_controls = AUDIO_CONTROLS_MASTER; continue;
#endif
				case 'p':
					switch (o = *p++) {
					case 'v':
						User_options.audio_pcm_volume = get_double(o, p, -1.0, 1.0);
						break;
					case 'b':
						User_options.audio_pcm_balance = get_double(o, p, -1.0, 1.0);
						break;
					default:
						FATAL(("illegal option: -ap%c", *(p - 1)));
						break;
					}
					break;
				case 'm':
					switch (o = *p++) {
					case 'v':
						User_options.audio_master_volume = get_double(o, p, -1.0, 1.0);
						break;
					case 'b':
						User_options.audio_master_balance = get_double(o, p, -1.0, 1.0);
						break;
					default:
						FATAL(("illegal option: -am%c", *(p - 1)));
						break;
					}
					break;
#ifndef LINUX
				case 'o':
					if (User_options.audio_ports == -1) {
						User_options.audio_ports = 0;
					}
					{
						int x = get_int(o, p, 0, (int)AUDIO_PORT_BIT_COUNT - 1);
						
						User_options.audio_ports |= (1 << x); break;
					}
					break;
#endif
				default:
					FATAL(("illegal option: -a%c", *(p - 1)));
					break;
				}
			}
			return;
		case 'e':	User_options.auto_exit = True; continue;

		case 'p':	User_options.auto_play = False; continue;

		case 'P':
			User_options.port = get_int(o, p, 1, (1 << BITS(short)) - 1);
			return;
		case 'U':
			/*
			 * UDP implementation is experimental at this point, and it does
			 * not work well.  It is not officially supported.
			 */ 
			User_options.port_udp = get_int(o, p, 1, (1 << BITS(short)) - 1);
			return;
		case 'd':
			User_options.dither_type = (DITHER_TYPE)get_int(o, p, 0, (int)DITHER_COUNT - 1);
			return;
		case 's':
			switch (o = *p) {
				case 'f':
					p++;
					User_options.framerate = get_double(o, p, FRAMERATE_MIN, FRAMERATE_MAX);
					break;
				default:
					User_options.sync_type = (SYNC_TYPE)get_int(o, p, 0, (int)SYNC_COUNT - 1);
					break;
			}
			return;

		case 'g':
			User_options.gamma = get_double(o, p, GAMMA_MIN, GAMMA_MAX);
			return;
		case 'c':
			User_options.contrast = get_double(o, p, CONTRAST_MIN, CONTRAST_MAX);
			return;
		case 'b':
			User_options.brightness = get_double(o, p, BRIGHTNESS_MIN, BRIGHTNESS_MAX);
			return;
		case 'x':
			Player_name = p;
			return;
		case 'W':	User_options.existing_window = get_int(o, p, 0, MAXINT);
					return;
		case 'S':	User_options.loop_start_pos = get_int(o, p, 0, MAXINT);
					return;
		case 'E':	User_options.loop_end_pos = get_int(o, p, 0, MAXINT);
					return;
		case '\0':	return;
		default:	FATAL(("unknown option -%c", o));
					EXIT(EXIT_FAILURE);
		}
	}
}

/*--------------------------------------------------------------------------+
| init_defaults																|
+--------------------------------------------------------------------------*/
static void
init_defaults()
{
	user_options_set_default();
}

/*--------------------------------------------------------------------------+
| main																		|
+--------------------------------------------------------------------------*/
/*ARGSUSED*/
int
main(argc, argv, envp)
	int argc;
	const char **argv, **envp;
{
	const char	**p, **q;
	int		n;

	init_signals();

	init_defaults();

	/* parse options */
	for (p = argv + 1, q = argv + argc, n = 0; p != q; p++) {
		if (**p == '-') {
			if ((*p)[1] == 0) {
				/* parameter '-' means read from stdin */
				User_options.stdin_mode = True;
				User_options.filename = NULL;
			}
		} else if (**p == '+') {
			/* all the options after a "mtv" are for xforms! */
			User_options.xforms_argv = p;
			break;
		} else {
			switch (++n) {
				case 1:		User_options.filename = *p; break;
				default:	FATAL(("too many parameters"));
			}
			continue;
		}
		parse_options(*p + 1);
	}

	if (User_options.filename != NULL || User_options.port != 0 || User_options.port_udp != 0 || User_options.stdin_mode) {
		if ((User_options.filename != NULL)+(User_options.port != 0)+(User_options.port_udp != 0)+(User_options.stdin_mode == True) != 1) {
			FATAL(("only one input should be specified"));
		}
	}

	Pid = getpid();

	MESSAGE(("started, pid = %u", (u_int)Pid));
	
	/*
	 * Start the player
	 */
	
	{
		char window_flag[12];
		char msg_flag[14];
		char audio_msg_flag[15];
		char audio_flag[4];
		const char *mtvp_argv[8];
		
		/*
		 * Set player options.
		 */
		
		window_flag[sizeof window_flag - 1] = 0;
		sprintf(window_flag, "-W%d", User_options.existing_window == -1 ? 0 : User_options.existing_window);
		if (window_flag[sizeof window_flag - 1] != 0) {
			abort_now();
		}
		msg_flag[sizeof msg_flag - 1] = 0;
		sprintf(msg_flag, "-w%d", - (User_options.msg_select + 1));
		if (msg_flag[sizeof msg_flag - 1] != 0) {
			abort_now();
		}
		audio_msg_flag[sizeof audio_msg_flag - 1] = 0;
		sprintf(audio_msg_flag, "-aw%d", - (User_options.audio_msg_select + 1));
		if (audio_msg_flag[sizeof audio_msg_flag - 1] != 0) {
			abort_now();
		}
		audio_flag[sizeof audio_flag - 1] = 0;
		sprintf(audio_flag, "-a%d", User_options.audio_mode);
		if (audio_flag[sizeof audio_flag - 1] != 0) {
			abort_now();
		}
		
		if (Player_name) {
			mtvp_argv[0] = Player_name;
		} else {
			mtvp_argv[0] = "mtvp";
		}
		mtvp_argv[1] = window_flag;
		mtvp_argv[2] = msg_flag;
		mtvp_argv[3] = audio_msg_flag;
		mtvp_argv[4] = audio_flag;
		mtvp_argv[5] = "-amv-1";	/* keep current master volume */
		mtvp_argv[6] = "-amb-1";	/* keep current master balance */
		mtvp_argv[7] = NULL;

		/*
		 * First parameter of g2a_init() is a ptr to a NULL-terminated array
		 * of strings.
		 * The first string is the name of the Player (e.g. "mtvp").
		 * The following strings are possible parameters.
		 * If the first string is a full pathname (starting with "/")
		 * g2a_init() will run this program as "the Player".
		 *
		 * If the first string is not a full pathname (starting with "/")
		 * g2a_init() will look for the Player in the following places:
		 * - in the directory where argv[0] is located (if argv[0] contains a '/')
		 * - in the current directory (".")
		 * - in the PATH
		 *
		 * Typically first string is "mtvp", if the player was installed
		 * under this name.
		 *
		 * Second parameter of g2a_init() is a string that should be
		 * the argv[0] from the calling program.
		 */
		 	
		if (g2a_init(&Player, mtvp_argv, argv[0]) == -1) {
			FATAL_ERROR(("Cannot fork player (%s) - check that player is in your PATH!", mtvp_argv[0]));
		}
	}
	
	/* do the work here */
	do_gui_fork(argv[0], User_options.xforms_argv);

	/* terminated successfully */
	EXIT(EXIT_SUCCESS);

	/* so that lint does not complain! */
	return 0;
}


/* eof */
