#ifndef lint
static char rcsid[] = "$Header: /home/tristan/mpeg/dec/RCS/gui_fork.c,v 1.82 1997/01/28 02:22:33 tristan Exp tristan $";
static char copyright[] = "Copyright (c) 1997, MpegTV LLC";
#endif

/*
 * Disclaimer of Warranty
 *
 * These software programs are available to the user without any license fee or
 * royalty on an "as is" basis.  Copyright-holder disclaims
 * any and all warranties, whether express, implied, or statuary, including any
 * implied warranties or merchantability or of fitness for a particular
 * purpose.  In no event shall the copyright-holder be liable for any
 * incidental, punitive, or consequential damages of any kind whatsoever
 * arising from the use of these programs.
 *
 * This disclaimer of warranty extends to the user of these programs and user's
 * customers, employees, agents, transferees, successors, and assigns.
 *
 */

/*
 * Motif interface:
 * Copyright (c) 1997, Alain Jobart (aj@bok.net)
 */


#include "misc.h"
#include "pmp.h"
#include "msg.h"
#include "exit.h"
#include "state_api.h"
#include "g2a.h"
#include "com_api.h"

#include <Xm/MainW.h>
#include <Xm/Form.h>
#include <Xm/PushB.h>
#include <Xm/FileSB.h>

/*--------------------------------------------------------------------------+
| Macros                                                                    |
+--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------+
| Data types                                                                |
+--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------+
| Globals                                                                    |
+--------------------------------------------------------------------------*/

/* motif variables */
Widget toplevel;
XtAppContext app_context;

Widget file_window = NULL;

/*--------------------------------------------------------------------------+
| Statics                                                                    |
+--------------------------------------------------------------------------*/

/* Default starting ressources of the application */
static char *app_resources[] =
{
    "*background:      grey",
    "*foreground:      black",
    "*text.background: whitesmoke",
    "*shadowThickness: 1",
    NULL
};

/* pid of this process */
static pid_t Pid;

/* to communicate with player */
static PMP_STRUCT    Player;

/*--------------------------------------------------------------------------+
| a2g_read_command                                                            |
+--------------------------------------------------------------------------*/
static void
a2g_read_command(cd, fd, id)
    XtPointer  cd;
    int        *fd;
    XtInputId  *id;
{
    PMP_HEADER    msg;
    PMP_DATA_A2G    data;
    
	/*
	 * This  callback is called when there is something available for
	 * reading on Player.receive_channel.msg_pipe[PMP_FD_READ]
	 * (also equal to fd in this routine).
	 *
	 * This routine should read one message from the Player
	 * and take the appropriate action.
	 *
	 * It is mostly a place holder in this example: This sample
	 * application ignores most messages.  It just checks the
	 * API version number and quits if the message A2G_COM_QUIT
	 * is received (which should not happen, by the way!).
	 */
	 
	/* read message from player */
	if (a2g_read_msg(&Player, &msg, &data) == -1) {
		FATAL_ERROR(("a2g_read_msg: error reading msg"));
	}

    MESSAGE(("a2g_read_command: received msg: %s", pmp_print_msg(&msg,  (const char *)&data)));
    
    switch ((A2G_COM)msg.type) {
		case A2G_COM_STATE:
			/* show Player's state in the form title */
			switch (data.data_generic.data_int) {
			case STATE_STOPPED:
				XtVaSetValues(toplevel, XmNtitle, "Simple Player (Stopped)", NULL);
				break;
			case STATE_PAUSED:
				XtVaSetValues(toplevel, XmNtitle, "Simple Player (Paused)", NULL);
				break;
			case STATE_PLAYING:
				XtVaSetValues(toplevel, XmNtitle, "Simple Player (Playing)", NULL);
				break;
			default:
				FATAL(("unknow State (should not happen!)"));
			}
			break;
        case A2G_COM_API_VERSION:
            if (data.data_generic.data_int != COM_API_VERSION) {
                FATAL(("API version mismatch, expecting %d, got %d. The mtvp program found in your PATH is not compatible with this front-end!", COM_API_VERSION, data.data_generic.data_int)); 
            }
            break;
        case A2G_COM_QUIT:
            FATAL(("Player decided to quit (should not happen!)"));
            break;
        default:
            /* msgs not understood should be ignored */
            break;
    }
}

/*--------------------------------------------------------------------------+
| OpenOkCB
+--------------------------------------------------------------------------*/
void OpenOkCB (Widget w, XtPointer cd, XtPointer call_data)
{
    /* get the file */
    XmFileSelectionBoxCallbackStruct *cbs =
        (XmFileSelectionBoxCallbackStruct *) call_data;
    char *filename = NULL;
    XmStringGetLtoR (cbs->value, XmSTRING_DEFAULT_CHARSET, &filename);

    /* do the regular stuff */
    g2a_send_string(&Player, COM_OPEN_FILE, 0, filename);

    /* and free the string */
    XtFree (filename);
    XtUnmanageChild (file_window);
}

/*--------------------------------------------------------------------------+
| OpenCancelCB
+--------------------------------------------------------------------------*/
void OpenCancelCB (Widget w, XtPointer cd, XtPointer cbs)
{
    XtUnmanageChild (file_window);
}

/*--------------------------------------------------------------------------+
| OpenCB
+--------------------------------------------------------------------------*/
void OpenCB (Widget w, XtPointer cd, XtPointer cbs)
{
    if (file_window == NULL) {
        Widget help_button;
        XmString t;

        file_window = XmCreateFileSelectionDialog (toplevel, "Open File",
                                                   NULL, 0);

        XtAddCallback (file_window, XmNokCallback, OpenOkCB, NULL);
        XtAddCallback (file_window, XmNcancelCallback, OpenCancelCB, NULL);

        /* get the help button and un-manage it */
        help_button = XtNameToWidget (file_window, "Help");
        if (help_button != NULL) XtUnmanageChild (help_button);

        /* set the file selection filter */
        t = XmStringCreateLocalized ("*.mpg");
        XmFileSelectionDoSearch (file_window, t);
        XmStringFree (t);

        /* and manage the box */
        XtManageChild (file_window);
        XtPopup (XtParent (file_window), XtGrabNone);
    } else {
        XtManageChild (file_window);
    }
}

/*--------------------------------------------------------------------------+
| PlayCB
+--------------------------------------------------------------------------*/
void PlayCB (Widget w, XtPointer cd, XtPointer cbs)
{
    g2a_send_int(&Player, COM_STATE, 0, STATE_PLAYING);
}

/*--------------------------------------------------------------------------+
| StepCB
+--------------------------------------------------------------------------*/
void StepCB (Widget w, XtPointer cd, XtPointer cbs)
{
    g2a_send_void(&Player, COM_STEP, 0);
}

/*--------------------------------------------------------------------------+
| RewindCB
+--------------------------------------------------------------------------*/
void RewindCB (Widget w, XtPointer cd, XtPointer cbs)
{
    g2a_send_int(&Player, COM_SEEK, 0, 0);
}

/*--------------------------------------------------------------------------+
| QuitCB
+--------------------------------------------------------------------------*/
void QuitCB (Widget w, XtPointer cd, XtPointer cbs)
{
    g2a_send_void(&Player, COM_QUIT, 0);
    EXIT(EXIT_SUCCESS);
}

/*--------------------------------------------------------------------------+
| init_form                                                                |
+--------------------------------------------------------------------------*/
static void
init_form(argc, argv)
    int        argc;
    const char **argv;
{    
    Widget main_w, main_form, w;

    /* Open the X connexion */
    XtSetLanguageProc (NULL, NULL, NULL);
    toplevel = XtVaAppInitialize (&app_context, "MpegTV Motif Front-end",
                                  NULL, 0, &argc, argv,
                                  app_resources, NULL, 0);

    main_w = XtVaCreateWidget ("main_w", xmMainWindowWidgetClass, toplevel,
                               NULL);
    
    main_form = XtVaCreateWidget ("form", xmFormWidgetClass, main_w,
                                  XmNfractionBase, 5,
                                  XmNhorizontalSpacing, 5,
                                  XmNverticalSpacing, 5,
                                  NULL);
    
    /* Create the open push button */
    w = XtVaCreateManagedWidget ("Open", xmPushButtonWidgetClass, main_form,
                                 XmNtopAttachment, XmATTACH_FORM,
                                 XmNbottomAttachment, XmATTACH_FORM,
                                 XmNleftAttachment, XmATTACH_FORM,
                                 XmNrightAttachment, XmATTACH_POSITION,
                                 XmNrightPosition, 1,
                                 XmNmarginHeight, 10,
                                 XmNmarginWidth, 20,
                                 NULL);
    XtAddCallback (w, XmNactivateCallback, OpenCB, NULL);

    /* Create the play push button */
    w = XtVaCreateManagedWidget ("Play", xmPushButtonWidgetClass, main_form,
                                 XmNtopAttachment, XmATTACH_FORM,
                                 XmNbottomAttachment, XmATTACH_FORM,
                                 XmNleftAttachment, XmATTACH_POSITION,
                                 XmNleftPosition, 1,
                                 XmNrightAttachment, XmATTACH_POSITION,
                                 XmNrightPosition, 2,
                                 NULL);
    XtAddCallback (w, XmNactivateCallback, PlayCB, NULL);

    /* Create the Step push button */
    w = XtVaCreateManagedWidget ("Step", xmPushButtonWidgetClass, main_form,
                                 XmNtopAttachment, XmATTACH_FORM,
                                 XmNbottomAttachment, XmATTACH_FORM,
                                 XmNleftAttachment, XmATTACH_POSITION,
                                 XmNleftPosition, 2,
                                 XmNrightAttachment, XmATTACH_POSITION,
                                 XmNrightPosition, 3,
                                 NULL);
    XtAddCallback (w, XmNactivateCallback, StepCB, NULL);

    /* Create the Rewind push button */
    w = XtVaCreateManagedWidget ("Rewind", xmPushButtonWidgetClass, main_form,
                                 XmNtopAttachment, XmATTACH_FORM,
                                 XmNbottomAttachment, XmATTACH_FORM,
                                 XmNleftAttachment, XmATTACH_POSITION,
                                 XmNleftPosition, 3,
                                 XmNrightAttachment, XmATTACH_POSITION,
                                 XmNrightPosition, 4,
                                 NULL);
    XtAddCallback (w, XmNactivateCallback, RewindCB, NULL);

    /* Create the Quit push button */
    w = XtVaCreateManagedWidget ("Quit", xmPushButtonWidgetClass, main_form,
                                 XmNtopAttachment, XmATTACH_FORM,
                                 XmNbottomAttachment, XmATTACH_FORM,
                                 XmNleftAttachment, XmATTACH_POSITION,
                                 XmNleftPosition, 4,
                                 XmNrightAttachment, XmATTACH_FORM,
                                 NULL);
    XtAddCallback (w, XmNactivateCallback, QuitCB, NULL);

    /* add the io callback */
    XtAppAddInput (app_context, Player.receive_channel.msg_pipe[PMP_FD_READ],
                   (XtPointer) XtInputReadMask, a2g_read_command, NULL);

    /* display the widgets */
    XtManageChild (main_form);
    XtManageChild (main_w);
    XtRealizeWidget (toplevel);
    
}


/*--------------------------------------------------------------------------+
| clean_exit                                                                |
+--------------------------------------------------------------------------*/
void
clean_exit(n)
    int        n;
{
    /* This is the routine called by the EXIT() macro */
    
    WARNING(("(pid=%u): exiting, status = %d", (u_int)Pid, n));

    /* kill the childs! */
    (void)signal(SIGCHLD, SIG_IGN);
    
    if (Player.pid != 0) {
        kill(Player.pid, SIGTERM);
    }
    
    /* terminate (real exit this time) */
    exit(n);
}


/*--------------------------------------------------------------------------+
| handler_sigchld                                                            |
+--------------------------------------------------------------------------*/
/*ARGSUSED*/
static void
handler_sigchld(sig)
    int     sig;
{
	/*
	 * Note: If you want to call any Xlib or Motif routine from
	 * a signal handler, you must use XtAppAddWorkProc()
	 * to handle signals in a Motif-based application, so that
	 * the signal handler will be executed in the proper context
	 * in the Motif event loop.
	 * Please refer to the Motif manual.
	 *
	 * This particular signal handler does not
	 * call any Xlib or Motif routine, so it is safe
	 * to install it with signal().
	 */

    WARNING(("(pid=%u): received signal %d", (u_int)Pid, sig));

    if (Player.pid != 0) {

		/* check if player actually died */
        switch (g2a_child_died(&Player)) {
        case 0:
            WARNING(("(pid=%u): child process %u still alive, signal ignored", (u_int)Pid, (u_int)Player.pid));
            break;
        case 1:
            WARNING(("(pid=%u): child process %u died, exiting...", (u_int)Pid, (u_int)Player.pid));
            EXIT(EXIT_FAILURE);
        case -1:
            FATAL_ERROR(("(pid=%u): g2a_child_died(%u) error", (u_int)Pid, (u_int)Player.pid));
        default:
            FATAL(("What?!? g2a_child_died() returned bad value!"));
        }            
    }

#ifdef SIGNAL_RESETS_HANDLER
    (void)signal(sig, handler_sigchld);
#endif
}


/*--------------------------------------------------------------------------+
| main                                                                        |
+--------------------------------------------------------------------------*/
/*ARGSUSED*/
int
main(argc, argv, envp)
    int argc;
    const char **argv, **envp;
{
    const char    **p, **q;
    int        n;

    /*
     * Set Msg_select to print all the warnings and messages.
     * Warning: if you set Msg_select before calling g2a_init(), you'll
     * also get all the messages in the player, since g2a_init() does
     * a fork and the player uses the same variable.
     */
     
    /* Msg_select = -1; */
    
    (void)signal(SIGCHLD, handler_sigchld);

    Pid = getpid();

    MESSAGE(("started, pid = %u", (u_int)Pid));
    
    /*
     * Start the player
     */
    
    {
        static const char *mtvp_argv[] = {
            "mtvp",    /* name of the player */
            /* "-w", */    /* possible options */
            NULL,    /* must be NULL-terminated */
        };

        /*
         * First parameter of g2a_init() is a ptr to a NULL-ternimated array
         * of strings.
         * The first string is the name of the Player ("mtvp").
         * The following strings are possible parameters.
         * If the first string is a full pathname (starting with "/")
         * g2a_init() will run this program as "the Player".
         *
         * If the first string is not a full pathname (starting with "/")
         * g2a_init() will look for the Player in the following places:
         * - in the directory where argv[0] is located (if argv[0] contains a '/')
         * - in the current directory (".")
         * - in the PATH
         *
         * Typically first string is "mtvp", if the player was installed
         * under this name.
         *
         * Second parameter of g2a_init() is a string that should be
         * the argv[0] from the calling program.
         */
    
        if (g2a_init(&Player, mtvp_argv, argv[0]) == -1) {
			FATAL_ERROR(("Cannot fork player - check that mtvp is in your PATH!"));
        }
    }
    
	/* initialize the form */
	init_form(argc, argv);
	
	/* if one argument was passed, assume that it is a file to open */
	if (argc == 2) {
		g2a_send_string(&Player, COM_OPEN_FILE, 0, argv[1]);
	} else if (argc > 2) {
		fprintf(stderr, "usage: simple_player_motif [file.mpg]\n");
		EXIT(EXIT_FAILURE);		
	}

    /* main event loop */
    MESSAGE(("entering event loop"));
    XtAppMainLoop (app_context);
    
    MESSAGE(("What?!? XtAppMainLoop() returned! (should never happen)"));

    /* terminated successfully */
    EXIT(EXIT_SUCCESS);

    /* so that lint does not complain! */
    return 0;
}


/* eof */
